from os import path
from time import time
import numpy as np
from silx.io.url import DataUrl
from ...resources.logger import LoggerOrPrint
from ...resources.utils import is_hdf5_extension, extract_parameters
from ...io.reader import ChunkReader, HDF5Loader, get_hdf5_dataset_shape
from ...preproc.ccd import Log, CCDFilter
from ...preproc.flatfield import FlatFieldDataUrls
from ...preproc.distortion import DistortionCorrection
from ...preproc.shift import VerticalShift
from ...preproc.double_flatfield import DoubleFlatField
from ...preproc.phase import PaganinPhaseRetrieval
from ...reconstruction.sinogram import SinoBuilder, SinoNormalization
from ...misc.rotation import Rotation
from ...preproc.rings import MunchDeringer
from ...misc.unsharp import UnsharpMask
from ...misc.histogram import PartialHistogram, hist_as_2Darray
from ..utils import use_options, pipeline_step, WriterConfigurator
# For now we don't have a plain python/numpy backend for reconstruction
try:
    from ...reconstruction.fbp_opencl import Backprojector
except:
    Backprojector = None


class ChunkedPipeline:
    """
    Pipeline for "regular" full-field tomography.
    Data is processed by chunks. A chunk consists in K contiguous lines of all the radios.
    In parallel geometry, a chunk of K radios lines gives K sinograms,
    and equivalently K reconstructed slices.
    """

    backend = "numpy"
    FlatFieldClass = FlatFieldDataUrls
    DoubleFlatFieldClass = DoubleFlatField
    CCDCorrectionClass = CCDFilter
    PaganinPhaseRetrievalClass = PaganinPhaseRetrieval
    UnsharpMaskClass = UnsharpMask
    ImageRotationClass = Rotation
    VerticalShiftClass = VerticalShift
    SinoBuilderClass = SinoBuilder
    SinoDeringerClass = MunchDeringer
    MLogClass = Log
    SinoNormalizationClass = SinoNormalization
    FBPClass = Backprojector
    HistogramClass = PartialHistogram

    def __init__(self, process_config, sub_region, logger=None, extra_options=None, phase_margin=None):
        """
        Initialize a "Chunked" pipeline.

        Parameters
        ----------
        processing_config: `ProcessConfig`
            Process configuration.
        sub_region: tuple
            Sub-region to process in the volume for this worker, in the format
            `(start_x, end_x, start_z, end_z)`.
        logger: `nabu.app.logger.Logger`, optional
            Logger class
        extra_options: dict, optional
            Advanced extra options.
        phase_margin: tuple, optional
            Margin to use when performing phase retrieval, in the form ((up, down), (left, right)).
            See also the documentation of PaganinPhaseRetrieval.
            If not provided, no margin is applied.


        Notes
        ------
        Using a `phase_margin` results in a lesser number of reconstructed slices.
        More specifically, if `phase_margin = (V, H)`, then there will be `delta_z - 2*V`
        reconstructed slices (if the sub-region is in the middle of the volume)
        or `delta_z - V` reconstructed slices (if the sub-region is on top or bottom
        of the volume).
        """
        self.logger = LoggerOrPrint(logger)
        self._set_params(process_config, sub_region, extra_options, phase_margin)
        self.set_subregion(sub_region)
        self._init_pipeline()


    def _set_params(self, process_config, sub_region, extra_options, phase_margin):
        self.process_config = process_config
        self.dataset_info = self.process_config.dataset_info
        self.dataset_infos = self.process_config.dataset_info # shorthand - deprecated
        self.processing_steps = self.process_config.processing_steps.copy()
        self.processing_options = self.process_config.processing_options
        self.sub_region = self._check_subregion(sub_region)
        self.delta_z = sub_region[-1] - sub_region[-2]
        self.chunk_size = self.delta_z
        self._set_phase_margin(phase_margin)
        self._set_extra_options(extra_options)
        self._callbacks = {}
        self._steps_name2component = {}
        self._steps_component2name = {}
        self._data_dump = {}
        self._resume_from_step = None


    @staticmethod
    def _check_subregion(sub_region):
        if len(sub_region) < 4:
            assert len(sub_region) == 2
            sub_region = (None, None) + sub_region
        if None in sub_region[-2:]:
            raise ValueError("Cannot set z_min or z_max to None")
        return sub_region


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {}
        advanced_options.update(extra_options)
        self.extra_options = advanced_options


    def _set_phase_margin(self, phase_margin):
        if phase_margin is None:
            phase_margin = ((0, 0), (0, 0))
        self._phase_margin_up = phase_margin[0][0]
        self._phase_margin_down = phase_margin[0][1]
        self._phase_margin_left = phase_margin[1][0]
        self._phase_margin_right = phase_margin[1][1]


    def set_subregion(self, sub_region):
        """
        Set a sub-region to process.

        Parameters
        ----------
        sub_region: tuple
            Sub-region to process in the volume, in the format
            `(start_x, end_x, start_z, end_z)` or `(start_z, end_z)`.
        """
        sub_region = self._check_subregion(sub_region)
        dz = sub_region[-1] - sub_region[-2]
        if dz != self.delta_z:
            raise ValueError(
                "Class was initialized for delta_z = %d but provided sub_region has delta_z = %d"
                % (self.delta_z, dz)
            )
        self.sub_region = sub_region
        self.z_min = sub_region[-2]
        self.z_max = sub_region[-1]


    def _compute_phase_kernel_margin(self):
        """
        Get the "margin" to pass to classes like PaganinPhaseRetrieval.
        In order to have a good accuracy for filter-based phase retrieval methods,
        we need to load extra data around the edges of each image. Otherwise,
        a default padding type is applied.
        """
        if not(self.use_radio_processing_margin):
            self._phase_margin = None
            return
        up_margin = self._phase_margin_up
        down_margin = self._phase_margin_down
        # Horizontal margin is not implemented
        left_margin, right_margin = (0, 0)
        self._phase_margin = ((up_margin, down_margin), (left_margin, right_margin))


    @property
    def use_radio_processing_margin(self):
        return ("phase" in self.processing_steps) or ("unsharp_mask" in self.processing_steps)


    def _get_phase_margin(self):
        if not(self.use_radio_processing_margin):
            return ((0, 0), (0, 0))
        return self._phase_margin


    def _get_cropped_radios(self):
        ((up_margin, down_margin), (left_margin, right_margin)) = self._phase_margin
        zslice = slice(up_margin or None, -down_margin or None)
        xslice = slice(left_margin or None, -right_margin or None)
        self._radios_cropped = self.radios[:, zslice, xslice]
        return self._radios_cropped


    @property
    def phase_margin(self):
        """
        Return the margin for phase retrieval in the form ((up, down), (left, right))
        """
        return self._get_phase_margin()


    @property
    def n_recs(self):
        """
        Return the final number of reconstructed slices.
        """
        n_recs = self.delta_z
        n_recs -= sum(self._get_phase_margin()[0])
        return n_recs


    def _get_process_name(self, kind="reconstruction"):
        # In the future, might be something like "reconstruction-<ID>"
        if kind == "reconstruction":
            return "reconstruction"
        elif kind == "histogram":
            return "histogram"
        return kind


    def _configure_dump(self, step_name):
        if step_name not in self.processing_steps:
            if step_name == "sinogram" and self.process_config._dump_sinogram:
                fname_full = self.process_config._dump_sinogram_file
            else:
                return
        else:
            if not self.processing_options[step_name].get("save", False):
                return
            fname_full = self.processing_options[step_name]["save_steps_file"]

        fname, ext = path.splitext(fname_full)
        dirname, file_prefix = path.split(fname)
        output_dir = path.join(dirname, file_prefix)
        file_prefix += str("_%04d" % self._get_image_start_index())

        self._data_dump[step_name] = WriterConfigurator(
            output_dir, file_prefix,
            file_format="hdf5",
            overwrite=True,
            logger=self.logger,
            nx_info = {
                "process_name": step_name,
                "processing_index": 0, # TODO
                "config": {
                    "processing_options": self.processing_options,
                    "nabu_config": self.process_config.nabu_config
                },
                "entry": getattr(self.dataset_info.dataset_scanner, "entry", None),
            }
        )


    def _configure_data_dumps(self):
        for step_name in self.processing_steps:
            self._configure_dump(step_name)
        # sinogram is a special keyword: not in processing_steps, but guaranteed to be before sinogram generation
        if self.process_config._dump_sinogram:
            self._configure_dump("sinogram")

    #
    # Callbacks
    #

    def register_callback(self, step_name, callback):
        """
        Register a callback for a pipeline processing step.

        Parameters
        ----------
        step_name: str
            processing step name
        callback: callable
            A function. It will be executed once the processing step `step_name`
            is finished. The function takes only one argument: the class instance.
        """
        if step_name not in self.processing_steps:
            raise ValueError(
                "'%s' is not in processing steps %s"
                % (step_name, self.processing_steps)
            )
        if step_name in self._callbacks:
            self._callbacks[step_name].append(callback)
        else:
            self._callbacks[step_name] = [callback]


    def _reshape_radios_after_phase(self):
        """
        Callback executed after phase retrieval, if margin != (0, 0).
        It modifies self.radios so that further processing will be done
        on the "inner part".
        """
        if sum(self._get_phase_margin()[0]) <= 0:
            return
        self._orig_radios = self.radios
        self.logger.debug(
            "Reshaping radios from %s to %s"
            % (str(self.radios.shape), str(self._radios_cropped.shape))
        )
        self.radios = self._radios_cropped

    #
    # Overwritten in inheriting classes
    #

    def _get_shape(self, step_name):
        """
        Get the shape to provide to the class corresponding to step_name.
        """
        if step_name == "flatfield":
            shape = self.radios.shape
        elif step_name == "double_flatfield":
            shape = self.radios.shape
        elif step_name == "rotate_projections":
            shape = self.radios.shape[1:]
        elif step_name == "phase":
            shape = self.radios.shape[1:]
        elif step_name == "ccd_correction":
            shape = self.radios.shape[1:]
        elif step_name == "unsharp_mask":
            shape = self.radios.shape[1:]
        elif step_name == "take_log":
            shape = self._radios_cropped_shape
        elif step_name == "radios_movements":
            shape = self._radios_cropped_shape
        elif step_name == "sino_normalization":
            shape = self._radios_cropped_shape
        elif step_name == "build_sino":
            shape = self._radios_cropped_shape
        elif step_name == "sino_rings_correction":
            shape = self.sino_builder.output_shape
        elif step_name == "reconstruction":
            shape = self.sino_builder.output_shape[1:]
        else:
            raise ValueError("Unknown processing step %s" % step_name)
        self.logger.debug("Data shape for %s is %s" % (step_name, str(shape)))
        return shape

    def _get_phase_output_shape(self):
        if not(self.use_radio_processing_margin):
            self._radios_cropped_shape = self.radios.shape
            return
        ((up_margin, down_margin), (left_margin, right_margin)) = self._phase_margin
        self._radios_cropped_shape = (
            self.radios.shape[0],
            self.radios.shape[1] - (up_margin + down_margin),
            self.radios.shape[2] - (left_margin + right_margin)
        )

    def _allocate_array(self, shape, dtype, name=None):
        return np.zeros(shape, dtype=dtype)

    def _allocate_sinobuilder_output(self):
        return self._allocate_array(self.sino_builder.output_shape, "f", name="sinos")

    def _allocate_recs(self, ny, nx):
        self.n_slices = self.radios.shape[1]  # TODO modify with vertical shifts
        if self.use_radio_processing_margin:
            self.n_slices -= sum(self.phase_margin[0])
        self.recs = self._allocate_array((self.n_slices, ny, nx), "f", name="recs")

    def _reset_memory(self):
        pass


    def _get_read_dump_subregion(self):
        read_opts = self.processing_options["read_chunk"]
        if read_opts.get("process_file", None) is None:
            return None
        dump_start_z, dump_end_z = read_opts["dump_start_z"], read_opts["dump_end_z"]
        relative_start_z = self.z_min - dump_start_z
        relative_end_z = relative_start_z + self.delta_z
        # (n_angles, n_z, n_x)
        subregion = (None, None, relative_start_z, relative_end_z, None, None)
        return subregion


    def _check_resume_from_step(self):
        if self._resume_from_step is None:
            return
        read_opts = self.processing_options["read_chunk"]
        expected_radios_shape = get_hdf5_dataset_shape(
            read_opts["process_file"],
            read_opts["process_h5_path"],
            sub_region=self._get_read_dump_subregion(),
        )
        # TODO check


    def _init_reader_finalize(self):
        """
        Method called after _init_reader
        """
        self._check_resume_from_step()
        self.radios = self.chunk_reader.data
        self._compute_phase_kernel_margin()
        self._get_phase_output_shape()


    def _process_finalize(self):
        """
        Method called once the pipeline has been executed
        """
        pass


    def _get_slice_start_index(self):
        return self.z_min + self._phase_margin_up

    _get_image_start_index = _get_slice_start_index

    #
    # Pipeline initialization
    #

    def _init_pipeline(self):
        self._init_reader()
        self._init_flatfield()
        self._init_double_flatfield()
        self._init_ccd_corrections()
        self._init_radios_rotation()
        self._init_phase()
        self._init_unsharp()
        self._init_radios_movements()
        self._init_mlog()
        self._init_sino_normalization()
        self._init_sino_builder()
        self._init_sino_rings_correction()
        self._prepare_reconstruction()
        self._init_reconstruction()
        self._init_histogram()
        self._init_writer()
        self._configure_data_dumps()

    @use_options("read_chunk", "chunk_reader")
    def _init_reader(self):
        if "read_chunk" not in self.processing_steps:
            raise ValueError("Cannot proceed without reading data")
        options = self.processing_options["read_chunk"]
        process_file = options.get("process_file", None)
        if process_file is None:
            # Standard case - start pipeline from raw data
            # ChunkReader always take a non-subsampled dictionary "files"
            self.chunk_reader = ChunkReader(
                options["files"],
                sub_region=self.sub_region,
                convert_float=True,
                binning=options["binning"],
                dataset_subsampling=options["dataset_subsampling"]
            )
        else:
            # Resume pipeline from dumped intermediate step
            self.chunk_reader = HDF5Loader(
                process_file,
                options["process_h5_path"],
                sub_region=self._get_read_dump_subregion()
            )
            self._resume_from_step = options["step_name"]
            self.logger.debug(
                "Load subregion %s from file %s" % (str(self.chunk_reader.sub_region), self.chunk_reader.fname)
            )
        self._init_reader_finalize()

    @use_options("flatfield", "flatfield")
    def _init_flatfield(self, shape=None):
        if shape is None:
            shape = self._get_shape("flatfield")
        options = self.processing_options["flatfield"]

        distortion_correction = None
        if options["do_flat_distortion"]:
            self.logger.info("Flats distortion correction will be applied")
            estimation_kwargs = {}
            estimation_kwargs.update(options["flat_distortion_params"])
            estimation_kwargs["logger"] = self.logger
            distortion_correction = DistortionCorrection(
                estimation_method="fft-correlation",
                estimation_kwargs=estimation_kwargs,
                correction_method="interpn"
            )

        # FlatField parameter "radios_indices" must account for subsampling
        self.flatfield = self.FlatFieldClass(
            shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=options["projs_indices"],
            interpolation="linear",
            distortion_correction=distortion_correction,
            sub_region=self.sub_region,
            binning=options["binning"],
            convert_float=True
        )

    @use_options("double_flatfield", "double_flatfield")
    def _init_double_flatfield(self):
        options = self.processing_options["double_flatfield"]
        avg_is_on_log = (options["sigma"] is not None)
        result_url = None
        if options["processes_file"] not in (None, ""):
            result_url = DataUrl(
                file_path=options["processes_file"],
                data_path=(self.dataset_info.hdf5_entry or "entry") + "/double_flatfield/results/data",
            )
            self.logger.info("Loading double flatfield from %s" % result_url.file_path())
        self.double_flatfield = self.DoubleFlatFieldClass(
            self._get_shape("double_flatfield"),
            result_url=result_url,
            sub_region=self.sub_region,
            input_is_mlog=False,
            output_is_mlog=False,
            average_is_on_log=avg_is_on_log,
            sigma_filter=options["sigma"]
        )

    @use_options("ccd_correction", "ccd_correction")
    def _init_ccd_corrections(self):
        options = self.processing_options["ccd_correction"]
        self.ccd_correction = self.CCDCorrectionClass(
            self._get_shape("ccd_correction"),
            median_clip_thresh=options["median_clip_thresh"]
        )

    @use_options("phase", "phase_retrieval")
    def _init_phase(self):
        options = self.processing_options["phase"]
        # If unsharp mask follows phase retrieval, then it should be done
        # before cropping to the "inner part".
        # Otherwise, crop the data just after phase retrieval.
        if "unsharp_mask" in self.processing_steps:
            margin = None
        else:
            margin=self._phase_margin
        self.phase_retrieval = self.PaganinPhaseRetrievalClass(
            self._get_shape("phase"),
            distance=options["distance_m"],
            energy=options["energy_kev"],
            delta_beta=options["delta_beta"],
            pixel_size=options["pixel_size_m"],
            padding=options["padding_type"],
            margin=margin,
            fftw_num_threads=False, # TODO tune in advanced params of nabu config file
        )
        if self.phase_retrieval.use_fftw:
            self.logger.debug(
                "PaganinPhaseRetrieval using FFTW with %d threads"
                % self.phase_retrieval.fftw.num_threads
            )
        if "unsharp_mask" not in self.processing_steps:
            self.register_callback("phase", ChunkedPipeline._reshape_radios_after_phase)

    @use_options("unsharp_mask", "unsharp_mask")
    def _init_unsharp(self):
        options = self.processing_options["unsharp_mask"]
        self.unsharp_mask = self.UnsharpMaskClass(
            self._get_shape("unsharp_mask"),
            options["unsharp_sigma"], options["unsharp_coeff"],
            mode="reflect", method=options["unsharp_method"]
        )
        self.register_callback("unsharp_mask", ChunkedPipeline._reshape_radios_after_phase)

    @use_options("take_log", "mlog")
    def _init_mlog(self):
        options = self.processing_options["take_log"]
        self.mlog = self.MLogClass(
            self._get_shape("take_log"),
            clip_min=options["log_min_clip"],
            clip_max=options["log_max_clip"]
        )

    @use_options("rotate_projections", "projs_rot")
    def _init_radios_rotation(self):
        options = self.processing_options["rotate_projections"]
        center = options["center"]
        if center is None:
            nx, ny = self.dataset_info.radio_dims
            center = (nx/2 - 0.5, ny/2 - 0.5)
        center = (center[0], center[1] - self.z_min)
        self.projs_rot = self.ImageRotationClass(
            self._get_shape("rotate_projections"),
            options["angle"],
            center=center,
            mode="edge",
            reshape=False
        )
        self._tmp_rotated_radio = self._allocate_array(
            self._get_shape("rotate_projections"), "f", name="tmp_rotated_radio"
        )

    @use_options("radios_movements", "radios_movements")
    def _init_radios_movements(self):
        options = self.processing_options["radios_movements"]
        self._vertical_shifts = options["translation_movements"][:, 1]
        self.radios_movements = self.VerticalShiftClass(
            self._get_shape("radios_movements"),
            self._vertical_shifts
        )

    @use_options("sino_normalization", "sino_normalization")
    def _init_sino_normalization(self):
        options = self.processing_options["sino_normalization"]
        self.sino_normalization = self.SinoNormalizationClass(
            kind=options["method"],
            radios_shape=self._get_shape("sino_normalization"),
        )

    @use_options("build_sino", "sino_builder")
    def _init_sino_builder(self):
        options = self.processing_options["build_sino"]
        self.sino_builder = self.SinoBuilderClass(
            radios_shape=self._get_shape("build_sino"),
            rot_center=options["rotation_axis_position"],
            halftomo=options["enable_halftomo"],
        )
        if not(options["enable_halftomo"]):
            self._sinobuilder_copy = False
            self._sinobuilder_output = None
            self.sinos = None
        else:
            self._sinobuilder_copy = True
            self.sinos = self._allocate_sinobuilder_output()
            self._sinobuilder_output = self.sinos

    @use_options("sino_rings_correction", "sino_deringer")
    def _init_sino_rings_correction(self):
        options = self.processing_options["sino_rings_correction"]
        fw_params = extract_parameters(options["user_options"])
        fw_sigma = fw_params.pop("sigma", 1.)
        self.sino_deringer = self.SinoDeringerClass(
            fw_sigma,
            sinos_shape=self._get_shape("sino_rings_correction"),
            **fw_params
        )

    # this should be renamed, as it could be confused with _init_reconstruction. What about _get_reconstruction_array ?
    @use_options("reconstruction", "reconstruction")
    def _prepare_reconstruction(self):
        options = self.processing_options["reconstruction"]
        x_s, x_e = options["start_x"], options["end_x"]
        y_s, y_e = options["start_y"], options["end_y"]
        self._rec_roi = (x_s, x_e + 1, y_s, y_e + 1)
        self._allocate_recs(y_e - y_s + 1, x_e - x_s + 1)


    @use_options("reconstruction", "reconstruction")
    def _init_reconstruction(self):
        options = self.processing_options["reconstruction"]
        if self.sino_builder is None:
            raise ValueError("Reconstruction cannot be done without build_sino")
        if self.FBPClass is None:
            raise ValueError("No usable FBP module was found")

        if options["enable_halftomo"]:
            rot_center = options["rotation_axis_position_halftomo"]
        else:
            rot_center = options["rotation_axis_position"]
        if self.sino_builder._halftomo_flip:
            rot_center = self.sino_builder.rot_center

        self.reconstruction = self.FBPClass(
            self._get_shape("reconstruction"),
            angles=options["angles"],
            rot_center=rot_center,
            filter_name=options["fbp_filter_type"],
            slice_roi=self._rec_roi,
            padding_mode=options["padding_type"],
            extra_options={
                "scale_factor": 1./options["pixel_size_cm"],
                "axis_correction": options["axis_correction"],
                "centered_axis": options["centered_axis"],
                "clip_outer_circle": options["clip_outer_circle"],
            }
        )
        if options["fbp_filter_type"] is None:
            self.reconstruction.fbp = self.reconstruction.backproj

    @use_options("histogram", "histogram")
    def _init_histogram(self):
        options = self.processing_options["histogram"]
        self.histogram = self.HistogramClass(
            method="fixed_bins_number", num_bins=options["histogram_bins"]
        )

    @use_options("save", "writer")
    def _init_writer(self):
        options = self.processing_options["save"]
        file_prefix = options["file_prefix"]
        output_dir = path.join(
            options["location"],
            file_prefix
        )
        nx_info = None
        self._hdf5_output = is_hdf5_extension(options["file_format"], errors=False)
        if self._hdf5_output:
            fname_start_index = None
            file_prefix += str("_%04d" % self._get_slice_start_index())
            entry = getattr(self.dataset_info.dataset_scanner, "entry", None)
            nx_info = {
                "process_name": self._get_process_name(),
                "processing_index": 0,
                "config": {
                    # "processing_options": self.processing_options, # Takes too much time to write, not useful for partial files
                    "nabu_config": self.process_config.nabu_config,
                },
                "entry": entry,
            }
            self._histogram_processing_index = nx_info["processing_index"] + 1
        else:
            fname_start_index = self._get_slice_start_index()
            self._histogram_processing_index = 1
        writer_options = {}
        if options["tiff_single_file"]:
            writer_options = {
                "tiff_single_file": options["tiff_single_file"],
                "single_tiff_initialized": getattr(self.process_config, "single_tiff_initialized", False),
            }
            self.process_config.single_tiff_initialized = True
        self._writer_configurator = WriterConfigurator(
            output_dir, file_prefix,
            file_format=options["file_format"],
            overwrite=options["overwrite"],
            start_index=fname_start_index,
            logger=self.logger,
            nx_info=nx_info,
            write_histogram=("histogram" in self.processing_steps),
            histogram_entry=getattr(self.dataset_info.dataset_scanner, "entry", "entry"),
            writer_options=writer_options,
            extra_options={
                "jpeg2000_compression_ratio": options["jpeg2000_compression_ratio"],
                "float_clip_values": options["float_clip_values"],
            }
        )
        self.writer = self._writer_configurator.writer
        self._writer_exec_args = self._writer_configurator._writer_exec_args
        self._writer_exec_kwargs = self._writer_configurator._writer_exec_kwargs
        self.histogram_writer = self._writer_configurator.get_histogram_writer()


    #
    # Pipeline re-initialization
    #

    def _reset_sub_region(self, sub_region):
        self.set_subregion(sub_region)
        self._reset_reader_subregion()
        self._reset_flatfield()

    def _reset_flatfield(self):
        self._init_flatfield()

    #
    # Pipeline execution
    #

    @pipeline_step("chunk_reader", "Reading data")
    def _read_data(self):
        self.logger.debug("Region = %s" % str(self.sub_region))
        t0 = time()
        self.chunk_reader.load_data()
        el = time() - t0

        shp = self.chunk_reader.data.shape
        itemsize = self.chunk_reader.dtype.itemsize if hasattr(self.chunk_reader, "dtype") else 4
        GB = np.prod(shp) * itemsize / 1e9
        self.logger.info(
            "Read subvolume %s (%.2f GB) in %.2f s: %.3f GB/s"
            % (str(shp), GB, el, GB/el)
        )

    def _reset_reader_subregion(self):
        if self._resume_from_step is None:
            self.chunk_reader._set_subregion(self.sub_region)
            self.chunk_reader._init_reader()
            self.chunk_reader._loaded = False
        else:
            self.chunk_reader._set_subregion(self._get_read_dump_subregion())
            self.chunk_reader._loaded = False


    @pipeline_step("flatfield", "Applying flat-field")
    def _flatfield(self):
        self.flatfield.normalize_radios(self.radios)

    @pipeline_step("double_flatfield", "Applying double flat-field")
    def _double_flatfield(self, radios=None):
        if radios is None:
            radios = self.radios
        self.double_flatfield.apply_double_flatfield(radios)

    @pipeline_step("ccd_correction", "Applying CCD corrections")
    def _ccd_corrections(self, radios=None):
        if radios is None:
            radios = self.radios
        _tmp_radio = self._allocate_array(radios.shape[1:], "f", name="tmp_ccdcorr_radio")
        for i in range(radios.shape[0]):
            self.ccd_correction.median_clip_correction(
                radios[i], output=_tmp_radio
            )
            radios[i][:] = _tmp_radio[:]

    @pipeline_step("projs_rot", "Rotating projections")
    def _rotate_projections(self, radios=None):
        if radios is None:
            radios = self.radios
        tmp_radio = self._tmp_rotated_radio
        for i in range(radios.shape[0]):
            self.projs_rot.rotate(radios[i], output=tmp_radio)
            radios[i][:] = tmp_radio[:]

    @pipeline_step("phase_retrieval", "Performing phase retrieval")
    def _retrieve_phase(self):
        if "unsharp_mask" in self.processing_steps:
            output = self.radios
        else:
            self._get_cropped_radios()
            output = self._radios_cropped
        for i in range(self.radios.shape[0]):
            self.phase_retrieval.apply_filter(
                self.radios[i], output=output[i]
            )

        # TEST
        # def retrieve_phase_func(radio_and_output):
        #     radio, output = radio_and_output
        #     self.phase_retrieval.apply_filter(
        #         radio, output=output
        #     )
        # from multiprocessing.pool import ThreadPool
        # with ThreadPool(10) as T:
        #     T.map(retrieve_phase_func, zip(self.radios, output))


    @pipeline_step("unsharp_mask", "Performing unsharp mask")
    def _apply_unsharp(self):
        for i in range(self.radios.shape[0]):
            self.radios[i] = self.unsharp_mask.unsharp(self.radios[i])
        self._get_cropped_radios()

    @pipeline_step("mlog", "Taking logarithm")
    def _take_log(self):
        self.mlog.take_logarithm(self.radios)

    @pipeline_step("radios_movements", "Applying radios movements")
    def _radios_movements(self, radios=None):
        if radios is None:
            radios = self.radios
        self.radios_movements.apply_vertical_shifts(
            radios, list(range(radios.shape[0]))
        )

    @pipeline_step("sino_normalization", "Normalizing sinograms")
    def _normalize_sinos(self, radios=None):
        if radios is None:
            radios = self.radios
        sinos = radios.transpose((1, 0, 2))
        self.sino_normalization.normalize(sinos)

    def _dump_sinogram(self, radios=None):
        if radios is None:
            radios = self.radios
        self._dump_data_to_file("sinogram", data=radios)

    @pipeline_step("sino_builder", "Building sinograms")
    def _build_sino(self, radios=None):
        if radios is None:
            radios = self.radios
        # Either a new array (previously allocated in "_sinobuilder_output"),
        # or a view of "radios"
        self.sinos = self.sino_builder.radios_to_sinos(
            radios,
            output=self._sinobuilder_output,
            copy=self._sinobuilder_copy
        )

    @pipeline_step("sino_deringer", "Removing rings on sinograms")
    def _destripe_sinos(self, sinos=None):
        if sinos is None:
            sinos = self.sinos
        self.sino_deringer.remove_rings(sinos)

    @pipeline_step("reconstruction", "Reconstruction")
    def _reconstruct(self, sinos=None):
        if sinos is None:
            sinos = self.sinos
        for i in range(sinos.shape[0]):
            self.reconstruction.fbp(
                sinos[i], output=self.recs[i]
            )

    @pipeline_step("histogram", "Computing histogram")
    def _compute_histogram(self, data=None):
        if data is None:
            data = self.recs
        self.recs_histogram = self.histogram.compute_histogram(data)

    @pipeline_step("writer", "Saving data")
    def _write_data(self, data=None):
        if data is None:
            data = self.recs
        self.writer.write(data, *self._writer_exec_args, **self._writer_exec_kwargs)
        self.logger.info("Wrote %s" % self.writer.get_filename())
        self._write_histogram()

    def _write_histogram(self):
        if "histogram" not in self.processing_steps:
            return
        self.logger.info("Saving histogram")
        self.histogram_writer.write(
            hist_as_2Darray(self.recs_histogram),
            self._get_process_name(kind="histogram"),
            processing_index=self._histogram_processing_index,
            config={
                "file": path.basename(self.writer.get_filename()),
                "bins": self.processing_options["histogram"]["histogram_bins"],
            }
        )

    def _dump_data_to_file(self, step_name, data=None):
        if step_name not in self._data_dump:
            return
        if data is None:
            data = self.radios
        writer = self._data_dump[step_name]
        self.logger.info("Dumping data to %s" % writer.fname)
        writer.write_data(data)


    def _process_chunk(self):
        self._flatfield()
        self._double_flatfield()
        self._ccd_corrections()
        self._rotate_projections()
        self._retrieve_phase()
        self._apply_unsharp()
        self._take_log()
        self._radios_movements()
        self._normalize_sinos()
        self._dump_sinogram()
        self._build_sino()
        self._destripe_sinos()
        self._reconstruct()
        self._compute_histogram()
        self._write_data()
        self._process_finalize()


    def process_chunk(self, sub_region=None):
        if sub_region is not None:
            self._reset_sub_region(sub_region)
            self._reset_memory()
            self._init_writer()
            self._init_double_flatfield()
            self._configure_data_dumps()
        self._read_data()
        self._process_chunk()



import numpy as np
import pytest
from nabu.utils import clip_circle
from nabu.testutils import get_data, compare_arrays
from nabu.reconstruction.rings import MunchDeringer
from nabu.thirdparty.pore3d_deringer_munch import  munchetal_filter
from nabu.cuda.utils import __has_pycuda__, get_cuda_context
__have_gpuderinger__ = False
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.reconstruction.rings_cuda import CudaMunchDeringer, __have_pycudwt__, __have_skcuda__
    if __have_pycudwt__ and __have_skcuda__:
        __have_gpuderinger__ = True


@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("mri_sino500.npz")["data"]
    cls.tol = 5e-3
    cls.rings = {
        150: 0.5,
        -150: 0.5
    }
    cls.fw_levels = 4
    cls.fw_sigma = 1.
    cls.fw_wname = "db5"
    if __have_gpuderinger__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)
    yield
    cls.ctx.pop()

@pytest.mark.usefixtures('bootstrap')
class TestMunchDeringer:

    @staticmethod
    def add_stripes_to_sino(sino, rings_desc):
        """
        Create a new sinogram by adding synthetic stripes to an existing one.

        Parameters
        ----------
        sino: array-like
            Sinogram.
        rings_desc: dict
            Dictionary describing the stripes locations and intensity.
            The location is an integer in [0, N[ where N is the number of columns.
            The intensity is a float: percentage of the current column mean value.
        """
        sino_out = np.copy(sino)
        for loc, intensity in rings_desc.items():
            sino_out[:, loc] += sino[:, loc].mean() * intensity
        return sino_out


    @pytest.mark.skipif(munchetal_filter is None, reason="Need PyWavelets for this test")
    def test_munch_deringer(self):
        deringer = MunchDeringer(
            self.fw_sigma,
            levels=self.fw_levels,
            wname=self.fw_wname,
            sinos_shape=self.sino.shape
        )
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        # Reference destriping with pore3d "munchetal_filter"
        ref = munchetal_filter(sino, self.fw_levels, self.fw_sigma, wname=self.fw_wname)
        # Wrapping with DeRinger
        res = np.zeros((1, ) + sino.shape, dtype=np.float32)
        deringer.remove_rings(sino, output=res)

        err_max = np.max(np.abs(res[0] - ref))
        assert err_max < self.tol, "Max error is too high"


    @pytest.mark.skipif(
        not(__have_gpuderinger__) or munchetal_filter is None,
        reason="Need pycuda, pycudwt and scikit-cuda for this test"
    )
    def test_cuda_munch_deringer(self):
        sino = self.add_stripes_to_sino(self.sino, self.rings)
        deringer = CudaMunchDeringer(
            self.fw_sigma, levels=self.fw_levels, wname=self.fw_wname, sinos_shape=self.sino.shape,
            cuda_options={"ctx": self.ctx}
        )
        d_sino = garray.to_gpu(sino)
        deringer.remove_rings(d_sino)
        res = d_sino.get()

        ref = munchetal_filter(sino, self.fw_levels, self.fw_sigma, wname=self.fw_wname)

        err_max = np.max(np.abs(res - ref))
        assert err_max < 1e-1, "Max error is too high"




import numpy as np
from ..utils import convert_index

class Reconstructor:
    """
    Abstract base class for reconstructors.

    A `Reconstructor` is a helper to reconstruct slices in arbitrary directions
    (not only usual "horizontal slices") in parallel-beam tomography.

    Current limitations:
       - Limitation to the three main axes
       - One instance of Reconstructor can only reconstruct successive slices

    Typical scenarios examples:
       - "I want to reconstruct several slices along 'z'", where `z` is the
         vertical axis. In this case, we reconstruct "horizontal slices" in planes
         perpendicular to the rotation axis.
       - "I want to reconstruct slices along 'y'". Here `y` is an axis perpendicular
         to `z`, i.e we reconstruct "vertical slices".

    A `Reconstructor` is tied to the set of slices to reconstruct (axis
    and orientation). Once defined, it cannot be changed ; i.e another class has
    to be instantiated to reconstruct slices in other axes/indices.

    The volume geometry conventions are defined below::


                              __________
                             /         /|
                            /         / |
             z             /         /  |
             ^            /_________/   |
             |           |          |   |
             |    y      |          |   /
             |   /       |          |  /
             |  /        |          | /
             | /         |__________|/
             |/
             ---------- > x


    The axis `z` parallel to the rotation axis.
    The usual parallel-beam tomography setting reconstructs slices along `z`,
    i.e in planes parallel to (x, y).
    """

    def __init__(self, shape, indices_range, axis="z", vol_type="sinograms", slices_roi=None):
        """
        Initialize a reconstructor.

        Parameters
        -----------
        shape: tuple
            Shape of the stack of sinograms or projections.
        indices_range: tuple
            Range of indices to reconstruct, in the form (start, end).
            As the standard Python behavior, the upper bound is not included.
            For example, to reconstruct 100 slices (numbered from 0 to 99),
            then you can provide (0, 100) or (0, None). Providing (0, 99) or (0, -1)
            will omit the last slice.
        axis: str
            Axis along which the slices are reconstructed. This axis is orthogonal
            to the slices planes. This parameter can be either "x", "y", or "z".
            Default is "z" (reconstruct slices perpendicular to the rotation axis).
        vol_type: str, optional
            Whether the parameter `shape` describes a volume of sinograms or
            projections. The two are the same except that axes 0 and 1
            are swapped. Can be "sinograms" (default) or "projections".
        slices_roi: tuple, optional
            Define a Region Of Interest to reconstruct a part of each slice.
            By default, the whole slice is reconstructed for each slice index.
            This parameter is in the form `(start_u, end_u, start_v, end_v)`,
            where `u` and `v` are horizontal and vertical axes on the reconstructed
            slice respectively, regardless of its orientation.
            If one of the values is set to None, it will be replaced by
            the corresponding default value.

        Examples
        ---------
        To reconstruct the first two horizontal slices, i.e along `z`:
          `R = Reconstructor(vol_shape, [0, 1])`
        To reconstruct vertical slices 0-100 along the `y` axis:
          `R = Reconstructor(vol_shape, (0, 100), axis="y")`
        """
        self._set_shape(shape, vol_type)
        self._set_axis(axis)
        self._set_indices(indices_range)
        self._configure_geometry(slices_roi)


    def _set_shape(self, shape, vol_type):
        if "sinogram" in vol_type.lower():
            self.vol_type = "sinograms"
        elif "projection" in vol_type.lower():
            self.vol_type = "projections"
        else:
            raise ValueError("vol_type can be either 'sinograms' or 'projections'")
        if len(shape) != 3:
            raise ValueError("Expected a 3D array description, but shape does not have 3 dims")
        self.shape = shape
        if self.vol_type == "sinograms":
            n_z, n_a, n_x = shape
        else:
            n_a, n_z, n_x = shape
        self.sinos_shape = (n_z, n_a, n_x)
        self.projs_shape = (n_a, n_z, n_x)
        self.data_shape = self.sinos_shape if self.vol_type == "sinograms" else self.projs_shape
        self.n_a = n_a
        self.n_x = n_x
        self.n_y = n_x # square slice by default
        self.n_z = n_z
        self._n = {"x": self.n_x, "y": self.n_y, "z": self.n_z}


    def _set_axis(self, axis):
        if axis.lower() not in ["x", "y", "z"]:
            raise ValueError("axis can be either 'x', 'y' or 'z' (got %s)" % axis)
        self.axis = axis


    def _set_indices(self, indices_range):
        start, end = indices_range
        npix = self._n[self.axis]
        start = convert_index(start, npix, 0)
        end = convert_index(end, npix, npix)
        self.indices_range = (start, end)
        self._idx_start = start
        self._idx_end = end
        self.indices = np.arange(start, end)


    def _configure_geometry(self, slices_roi):
        self.slices_roi = slices_roi or (None, None, None, None)
        start_u, end_u, start_v, end_v = self.slices_roi
        uv_to_xyz = {
            "z": ("x", "y"), # reconstruct along z: u = x, v = y
            "y": ("y", "z"), # reconstruct along y: u = y, v = z
            "x": ("y", "z"), # reconstruct along x: u = y, v = z
        }
        rotated_axes = uv_to_xyz[self.axis]
        u_max = self._n[rotated_axes[0]]
        v_max = self._n[rotated_axes[1]]
        start_u = convert_index(start_u, u_max, 0)
        end_u = convert_index(end_u, u_max, u_max)
        start_v = convert_index(start_v, v_max, 0)
        end_v = convert_index(end_v, v_max, v_max)
        self.slices_roi = (start_u, end_u, start_v, end_v)

        if self.axis == "z":
            self.backprojector_roi = self.slices_roi
            start_z, end_z = self._idx_start, self._idx_end
        if self.axis == "y":
            self.backprojector_roi = (start_u, end_u, self._idx_start, self._idx_end)
            start_z, end_z = start_v, end_v
        if self.axis == "x":
            self.backprojector_roi = (self._idx_start, self._idx_end, start_u, end_u)
            start_z, end_z = start_v, end_v
        self._z_indices = np.arange(start_z, end_z)
        self.output_shape = (
            self._z_indices.size,
            self.backprojector_roi[3] - self.backprojector_roi[2],
            self.backprojector_roi[1] - self.backprojector_roi[0],
        )


    def _check_data(self, data):
        if data.shape != self.data_shape:
            raise ValueError(
                "Invalid data shape: expected %s shape %s, but got %s"
                % (self.vol_type, self.data_shape, data.shape)
            )
            if data.dtype != np.float32:
                raise ValueError("Expected float32 data type")


    def reconstruct(self):
        raise ValueError("Base class")


    __call__ = reconstruct



import os
from ..dataset_validator import DatasetValidatorBase

class FullFieldDatasetValidator(DatasetValidatorBase):

    def _validate(self):
        self._check_not_empty()
        self._convert_negative_indices()
        self._get_resources()
        self._get_output_filename()
        self._check_can_do_flatfield()
        self._check_slice_indices()
        self._handle_processing_mode()
        self._handle_binning()
        self._check_output_file()


    def _check_can_do_flatfield(self):
        if self.nabu_config["preproc"]["flatfield"]:
            darks = self.dataset_info.darks
            assert len(darks) > 0, "Need at least one dark to perform flat-field correction"
            for dark_id, dark_url in darks.items():
                assert os.path.isfile(dark_url.file_path()), "Dark file %s not found" % dark_url.file_path()
            flats = self.dataset_info.flats
            assert len(flats) > 0, "Need at least one flat to perform flat-field correction"
            for flat_id, flat_url in flats.items():
                assert os.path.isfile(flat_url.file_path()), "Flat file %s not found" % flat_url.file_path()


    def _check_slice_indices(self):
        nx, nz = self.dataset_info.radio_dims
        rec_params = self.nabu_config["reconstruction"]
        if rec_params["enable_halftomo"]:
            ny, nx = self._get_nx_ny()
        what = (
            ("start_x", "end_x", nx),
            ("start_y", "end_y", nx),
            ("start_z", "end_z", nz)
        )
        for (start_name, end_name, numels) in what:
            self._check_start_end_idx(
                rec_params[start_name],
                rec_params[end_name],
                numels,
                start_name=start_name,
                end_name=end_name
            )



from typing import Union
import numpy as np
from ..preproc.ccd import CCDFilter, Log
from ..cuda.kernel import CudaKernel
from ..cuda.medfilt import MedianFilter
from ..utils import get_cuda_srcfile, updiv, deprecated_class

# COMPAT.
from .flatfield_cuda import (
    CudaFlatField as CudaFlatfield_,
    CudaFlatFieldArrays as CudaFlatFieldArrays_,
    CudaFlatFieldDataUrls as CudaFlatFieldDataUrls_,
)

FlatField = deprecated_class(
    "preproc.ccd_cuda.CudaFlatField was moved to preproc.flatfield_cuda.CudaFlatField", do_print=True
)(CudaFlatfield_)
FlatFieldArrays = deprecated_class(
    "preproc.ccd_cuda.CudaFlatFieldArrays was moved to preproc.flatfield_cuda.CudaFlatFieldArrays", do_print=True
)(CudaFlatFieldArrays_)
FlatFieldDataUrls = deprecated_class(
    "preproc.ccd_cuda.CudaFlatFieldDataUrls was moved to preproc.flatfield_cuda.CudaFlatFieldDataUrls", do_print=True
)(CudaFlatFieldDataUrls_)
#


class CudaCCDFilter(CCDFilter):
    def __init__(
        self,
        radios_shape: tuple,
        correction_type: str = "median_clip",
        median_clip_thresh: float = 0.1,
        abs_diff=False,
        cuda_options: Union[dict, None] = None,
    ):
        """
        Initialize a CudaCCDCorrection instance.
        Please refer to the documentation of CCDCorrection.
        """
        super().__init__(
            radios_shape,
            correction_type=correction_type,
            median_clip_thresh=median_clip_thresh,
        )
        self._set_cuda_options(cuda_options)
        self.cuda_median_filter = None
        if correction_type == "median_clip":
            self.cuda_median_filter = MedianFilter(
                self.shape,
                footprint=(3, 3),
                mode="reflect",
                threshold=median_clip_thresh,
                abs_diff=abs_diff,
                cuda_options={
                    "device_id": self.cuda_options["device_id"],
                    "ctx": self.cuda_options["ctx"],
                    "cleanup_at_exit": self.cuda_options["cleanup_at_exit"],
                }
            )

    def _set_cuda_options(self, user_cuda_options):
        self.cuda_options = {"device_id": None, "ctx": None, "cleanup_at_exit": None}
        if user_cuda_options is None:
            user_cuda_options = {}
        self.cuda_options.update(user_cuda_options)

    def median_clip_correction(self, radio, output=None):
        """
        Compute the median clip correction on one image.

        Parameters
        ----------
        radio: pycuda.gpuarray
            A radio image
        output: pycuda.gpuarray, optional
            Output data.
        """
        assert radio.shape == self.shape
        return self.cuda_median_filter.medfilt2(radio, output=output)


CudaCCDCorrection = deprecated_class("CudaCCDCorrection is replaced with CudaCCDFilter", do_print=True)(CudaCCDFilter)


class CudaLog(Log):
    """
    Helper class to take -log(radios)
    """

    def __init__(self, radios_shape, clip_min=None, clip_max=None):
        """
        Initialize a Log processing.

        Parameters
        -----------
        radios_shape: tuple
            The shape of 3D radios stack.
        clip_min: float, optional
            Data smaller than this value is replaced by this value.
        clip_max: float, optional.
            Data bigger than this value is replaced by this value.
        """
        super().__init__(radios_shape, clip_min=clip_min, clip_max=clip_max)
        self._init_kernels()

    def _init_kernels(self):
        self._do_clip_min = int(self.clip_min is not None)
        self._do_clip_max = int(self.clip_max is not None)
        self.clip_min = np.float32(self.clip_min or 0)
        self.clip_max = np.float32(self.clip_max or 1)
        self._nlog_srcfile = get_cuda_srcfile("ElementOp.cu")
        nz, ny, nx = self.radios_shape
        self._nx = np.int32(nx)
        self._ny = np.int32(ny)
        self._nz = np.int32(nz)
        self._nthreadsperblock = (16, 16, 4)  # TODO tune ?
        self._nblocks = tuple([updiv(n, p) for n, p in zip([nx, ny, nz], self._nthreadsperblock)])

        self.nlog_kernel = CudaKernel(
            "nlog",
            filename=self._nlog_srcfile,
            signature="Piiiff",
            options=[
                "-DDO_CLIP_MIN=%d" % self._do_clip_min,
                "-DDO_CLIP_MAX=%d" % self._do_clip_max,
            ],
        )

    def take_logarithm(self, radios, clip_min=None, clip_max=None):
        """
        Take the negative logarithm of a radios chunk.

        Parameters
        -----------
        radios: `pycuda.gpuarray.GPUArray`
            Radios chunk
            If not provided, a new GPU array is created.
        clip_min: float, optional
            Before taking the logarithm, the values are clipped to this minimum.
        clip_max: float, optional
            Before taking the logarithm, the values are clipped to this maximum.
        """
        clip_min = clip_min or self.clip_min
        clip_max = clip_max or self.clip_max
        if radios.flags.c_contiguous:
            self.nlog_kernel(
                radios,
                self._nx,
                self._ny,
                self._nz,
                clip_min,
                clip_max,
                grid=self._nblocks,
                block=self._nthreadsperblock,
            )
        else:
            # map-like operations cannot be directly applied on 3D arrays
            # that are not C-contiguous. We have to process image per image.
            nz = np.int32(1)
            nthreadsperblock = (32, 32, 1)
            nblocks = tuple([updiv(n, p) for n, p in zip([int(self._nx), int(self._ny), int(nz)], nthreadsperblock)])
            for i in range(radios.shape[0]):
                self.nlog_kernel(
                    radios[i], self._nx, self._ny, nz, clip_min, clip_max, grid=nblocks, block=nthreadsperblock
                )
        return radios



import logging
import logging.config

class Logger(object):
    def __init__(self, loggername, level="DEBUG", logfile="logger.log", console=True):
        """
        Configure a Logger object.

        Parameters
        -----------
        loggername: str
            Logger name.
        level: str, optional
            Logging level. Can be "debug", "info", "warning", "error", "critical".
            Default is "debug".
        logfile: str, optional
            File where the logs are written.
            If set to None, the logs are not written in a file.
            Default is "logger.log".
        console: bool, optional
            If set to True, the logs are (also) written in stdout/stderr.
            Default is True.
        """
        self.loggername = loggername
        self.level = level
        self.logfile = logfile
        self.console = console
        self._configure_logger()

    def _configure_logger(self):
        conf = self._get_default_config_dict()
        for handler in conf["handlers"].keys():
            conf["handlers"][handler]["level"] = self.level.upper()
        conf["loggers"][self.loggername]["level"] = self.level.upper()
        if not(self.console):
            conf["loggers"][self.loggername]["handlers"].remove("console")

        self.config = conf
        logging.config.dictConfig(conf)
        self.logger = logging.getLogger(self.loggername)

    def _get_default_config_dict(self):
        conf = {
            'version': 1,
            'formatters': {
                'default': {'format': '%(asctime)s - %(levelname)s - %(message)s', 'datefmt': '%d-%m-%Y %H:%M:%S'},
                'console': {'format': '%(message)s'}
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'console',
                    'stream': 'ext://sys.stdout'
                },
                'file': {
                    'level': 'DEBUG',
                    'class': 'logging.FileHandler',
                    'formatter': 'default',
                    'filename': self.logfile,
                }
            },
            'loggers': {
                self.loggername: {
                    'level': 'DEBUG',
                    'handlers': ['console', 'file'],
                    # This logger inherits from root logger - dont propagate !
                    'propagate': False,
                }
            },
            'disable_existing_loggers': False,
        }
        return conf

    def info(self, msg):
        return self.logger.info(msg)

    def debug(self, msg):
        return self.logger.debug(msg)

    def warning(self, msg):
        return self.logger.warning(msg)

    warn = warning

    def error(self, msg):
        return self.logger.error(msg)

    def fatal(self, msg):
        return self.logger.fatal(msg)

    def critical(self, msg):
        return self.logger.critical(msg)



def LoggerOrPrint(logger):
    """
    Logger that is either a "true" logger object, or a fall-back to "print".
    """
    if logger is None:
        return PrinterLogger()
    return logger


class PrinterLogger(object):
    def __init__(self):
        methods = [
            "debug",
            "warn",
            "warning",
            "info",
            "error",
            "fatal",
            "critical",
        ]
        for method in methods:
            self.__setattr__(
                method,
                print
            )


LogLevel = {
    "notset": logging.NOTSET,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARN,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "fatal": logging.FATAL
}



import os
import numpy as np
from tomoscan.io import HDF5File
from silx.third_party.EdfFile import EdfFile
from ..io.utils import get_first_hdf5_entry
from .utils import parse_params_values
from .cli_configs import GenerateInfoConfig


edf_header_hdf5_path = {
    # EDF Header
    "count_time": "/technique/scan/exposure_time", # ms in HDF5
    "date": "/start_time",
    "energy": "/technique/scan/energy",
    "flip": "/technique/detector/flipping", # [x, y] in HDF5
    "motors": "/instrument/positioners",
    "optic_used": "/technique/optic/magnification",
}

info_hdf5_path = {
    "Energy": "/technique/scan/energy",
    "Distance": "/technique/scan/sample_detector_distance", # mm in HDF5 and EDF
    # ~ "Prefix": TODO
    "Directory": "/technique/saving/path",
    "ScanRange": "/technique/scan/scan_range",
    "TOMO_N": "/technique/scan/tomo_n",
    "REF_ON": "/technique/scan/ref_on",
    "REF_N": "/technique/reference/ref_n",
    "DARK_N": "/technique/dark/dark_n",
    "Dim_1": "/technique/detector/size", # array
    "Dim_2": "/technique/detector/size", # array
    "Count_time": "/technique/scan/exposure_time", # ms in HDF5, s in EDF ?
    "Shutter_time": "/technique/scan/shutter_time", # ms in HDF5, s in EDF ?
    "Optic_used": "/technique/optic/magnification",
    "PixelSize": "/technique/detector/pixel_size", # array (microns)
    "Date": "/start_time",
    "Scan_Type": "/technique/scan/scan_type",
    "SrCurrent": "/instrument/machine/current",
    "Comment": "/technique/scan/comment", # empty ?
}


def decode_bytes(content):
    if isinstance(content, bytes):
        return content.decode()
    else:
        return content


def simulate_edf_header(fname, entry, return_dict=False):
    edf_header = {
        "ByteOrder": "LowByteFirst",
    }
    with HDF5File(fname, "r") as fid:
        for edf_name, h5_path in edf_header_hdf5_path.items():
            if edf_name == "motors": continue
            h5_path = entry + h5_path
            edf_header[edf_name] = decode_bytes(fid[h5_path][()])
        h5_motors = decode_bytes(fid[entry + edf_header_hdf5_path["motors"]])
        edf_header["motor_mne"] = list(h5_motors.keys())
        edf_header["motor_pos"] = [v[()] for v in h5_motors.values()]

    # remove "scan_numbers" from motors
    try:
        idx = edf_header["motor_mne"].index("scan_numbers")
        edf_header["motor_mne"].pop(idx)
        edf_header["motor_pos"].pop(idx)
    except ValueError:
        pass

    # remove invalid values from motor_pos
    invalid_values = ["*DIS*"]
    try:
        for invalid_val in invalid_values:
            idx = edf_header["motor_pos"].index(invalid_val)
            edf_header["motor_mne"].pop(idx)
            edf_header["motor_pos"].pop(idx)
    except ValueError:
        pass

    # Format
    edf_header["flip"] = "<flip x : %s,flip y : %s>" % (edf_header["flip"][0], edf_header["flip"][1])
    edf_header["motor_mne"] = " ".join(edf_header["motor_mne"])
    edf_header["motor_pos"] = " ".join(list(map(str, edf_header["motor_pos"])))
    edf_header["count_time"] = edf_header["count_time"] * 1e-3 # HDF5: ms -> EDF: s

    if return_dict:
        return edf_header
    res = ""
    for k, v in edf_header.items():
        res = res + "%s = %s ;\n" % (k, v)
    return res


def simulate_info_file(fname, entry, return_dict=False):
    info_file_content = {}
    with HDF5File(fname, "r") as fid:
        for info_name, h5_path in info_hdf5_path.items():
            h5_path = entry + h5_path
            info_file_content[info_name] = decode_bytes(fid[h5_path][()])

    info_file_content["Dim_1"] = info_file_content["Dim_1"][0]
    info_file_content["Dim_2"] = info_file_content["Dim_2"][1]
    info_file_content["PixelSize"] = info_file_content["PixelSize"][0]
    info_file_content["Prefix"] = os.path.basename(fname)
    info_file_content["Col_end"] = info_file_content["Dim_1"] - 1
    info_file_content["Col_beg"] = 0
    info_file_content["Row_end"] = info_file_content["Dim_2"] - 1
    info_file_content["Row_beg"] = 0
    for what in ["Count_time", "Shutter_time"]:
        info_file_content[what] = info_file_content[what] * 1e-3

    if return_dict:
        return info_file_content
    # Format
    res = ""
    for k, v in info_file_content.items():
        k_s = str("%s=" % k)
        sep = " " * (24 - len(k_s))
        res = res + k_s + sep + str("%s\n" % v)
    return res



def get_hst_saturations(hist, bins, numels):
    aMin = bins[0]
    aMax = bins[-1]
    hist_sum = numels * 1.

    hist_cum = np.cumsum(hist)
    hist_cum_rev = np.cumsum(hist[::-1])

    i_s1 = np.where(hist_cum > 0.00001*hist_sum)[0][0]
    sat1 = aMin+i_s1*(aMax-aMin)/(hist.size - 1)

    i_S1 = np.where(hist_cum > 0.002*hist_sum)[0][0]
    Sat1 = aMin+i_S1*(aMax-aMin)/(hist.size-1)

    i_s2 = np.argwhere(hist_cum_rev > 0.00001*hist_sum)[0][0]
    sat2 = aMin+(hist.size - 1 - i_s2)*(aMax-aMin)/(hist.size - 1)

    i_S2 = np.argwhere(hist_cum_rev > 0.002*hist_sum)[0][0]
    Sat2 = aMin+(hist.size - 1 - i_S2)*(aMax-aMin)/(hist.size-1)

    return sat1, sat2, Sat1, Sat2


def simulate_hst_vol_header(fname, entry=None, return_dict=False):
    with HDF5File(fname, "r") as fid:
        try:
            histogram_path = entry + "/histogram/results/data"
            histogram_config_path = entry + "/histogram/configuration"
            Nz, Ny, Nx = fid[histogram_config_path]["volume_shape"][()]
            vol_header_content = {
                "NUM_X": Nx,
                "NUM_Y": Ny,
                "NUM_Z": Nz,
                "BYTEORDER": "LOWBYTEFIRST",
            }
            hist = fid[histogram_path][()]
        except KeyError as err:
            print("Could not load histogram from %s: %s" % (fname, err))
            return {} if return_dict else ""
    bins = hist[1]
    hist = hist[0]
    vmin = bins[0]
    vmax = bins[-1] + (bins[-1] - bins[-2])
    s1, s2, S1, S2 = get_hst_saturations(hist, bins, Nx*Ny*Nz)
    vmin = bins[0]
    vmax = bins[-1] + (bins[-1] - bins[-2])
    vol_header_content.update({
        "ValMin": vmin,
        "ValMax": vmax,
        "s1": s1,
        "s2": s2,
        "S1": S1,
        "S2": S2,
    })
    if return_dict:
        return vol_header_content
    res = ""
    for k, v in vol_header_content.items():
        res = res + "%s =  %s\n" % (k, v)
    return res


def format_as_info(d):
    res = ""
    for k, v in d.items():
        k_s = str("%s=" % k)
        sep = " " * (24 - len(k_s))
        res = res + k_s + sep + str("%s\n" % v)
    return res


def generate_merged_info_file_content(
    hist_fname=None, hist_entry=None,
    bliss_fname=None, bliss_entry=None,
    first_edf_proj=None, info_file=None
):
    # EDF Header
    if first_edf_proj is None:
        edf_header = simulate_edf_header(
            bliss_fname, bliss_entry, return_dict=True
        )
    else:
        edf_header = EdfFile(first_edf_proj).GetHeader(0)
    # .info File
    if info_file is None:
        info_file_content = simulate_info_file(
            bliss_fname, bliss_entry, return_dict=True
        )
        info_file_content = format_as_info(info_file_content)
    else:
        with open(info_file, "r") as f:
            info_file_content = f.read()
    # .vol File
    vol_file_content = simulate_hst_vol_header(
        hist_fname, entry=hist_entry, return_dict=True
    )
    #
    res = format_as_info(edf_header)
    res += info_file_content
    res += format_as_info(vol_file_content)
    return res



def str_or_none(s):
    if len(s.strip()) == 0:
        return None
    return s



def generate_merged_info_file():
    args = parse_params_values(
        GenerateInfoConfig,
        parser_description="Generate a .info file"
    )

    hist_fname = str_or_none(args["hist_file"])
    hist_entry = str_or_none(args["hist_entry"])
    bliss_fname = str_or_none(args["bliss_file"])
    bliss_entry = str_or_none(args["bliss_entry"])
    first_edf_proj = str_or_none(args["edf_proj"])
    info_file = str_or_none(args["info_file"])
    if hist_fname is not None and hist_entry is None:
        hist_entry = get_first_hdf5_entry(hist_fname)
    if bliss_fname is not None and bliss_entry is None:
        bliss_entry = get_first_hdf5_entry(bliss_fname)


    if not((bliss_fname is None) ^ (info_file is None)):
        print("Error: please provide either --bliss_file or --info_file")
        exit(1)

    if info_file is not None and first_edf_proj is None:
        print("Error: please provide also --edf_proj when using the EDF format")
        exit(1)

    content = generate_merged_info_file_content(
        hist_fname=hist_fname,
        hist_entry=hist_entry,
        bliss_fname=bliss_fname,
        bliss_entry=bliss_entry,
        first_edf_proj=first_edf_proj,
        info_file=info_file
    )
    with open(args["output"], "w") as f:
        f.write(content)





"""
Database for describing standard servers
"""

ESRF_HPC_legacy_gpu = {
    "memory_GB": 128.,
    "gpus": [
        {
            "type": "cuda",
            "name": "Tesla K20m",
            "memory_GB": 4.9,
            "compute_capability": (3, 5),
        },
        {
            "type": "cuda",
            "name": "Tesla K20m",
            "memory_GB": 4.9,
            "compute_capability": (3, 5),
        },
    ],
    "cpu": {
        "name": "Intel(R) Xeon(R) CPU E5-2670 0",
        "threads": 16,
        "sockets": 2,
        "threads_per_core": 1,
    },
}

ESRF_HPC_legacy_cpu = {
    "memory_GB": 256.,
    "gpus": [],
    "cpu": {
        "name": "Intel(R) Xeon(R) CPU E5-2680 v4",
        "threads": 28,
        "sockets": 2,
        "threads_per_core": 1,
    },
}


IBM_AC922 = {
    "memory_GB": 616.0,
    "gpus": [
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
    ],
    "cpu": {
        "name": "POWER9",
        "threads": 128,
        "sockets": 2,
        "threads_per_core": 4,
    },
}

ESRF_LBS191 = {
    "memory_GB": 540.0,
    "gpus": [
        {
            "type": "cuda",
            "name": "GeForce GTX 1080 Ti",
            "memory_GB": 11.7,
            "compute_capability": (6, 1),
        },
    ],
    "cpu": {
        "name": "Intel(R) Xeon(R) CPU E5-2697 v3",
        "threads": 56,
        "sockets": 2,
        "threads_per_core": 2,
    },
}

DGX1 = {
    "memory_GB": 540.,
    "gpus": [
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
        {
            "type": "cuda",
            "name": "Tesla V100-SXM2-32GB",
            "memory_GB": 33.8,
            "compute_capability": (7, 0),
        },
    ],
    "cpu": {
        "name": "Intel(R) Xeon(R) CPU E5-2698 v4",
        "threads": 80,
        "sockets": 2,
        "threads_per_core": 2,
    },
}



from os.path import join, isfile, basename, dirname
from math import ceil
import gc
from psutil import virtual_memory
from silx.io import get_data
from silx.io.url import DataUrl
from ...utils import remove_items_from_list, restore_items_in_list
from ...resources.logger import LoggerOrPrint
from ...io.writer import merge_hdf5_files, NXProcessWriter
from ...cuda.utils import collect_cuda_gpus
from ...preproc.phase import compute_paganin_margin
from ...misc.histogram import PartialHistogram, add_last_bin, hist_as_2Darray
from .chunked_cuda import CudaChunkedPipeline, CudaChunkedPipelineLimitedMemory
from .grouped import GroupedPipeline, SinoStackPipeline
from .grouped_cuda import CudaGroupedPipeline, CudaSinoStackPipeline
from .computations import estimate_chunk_size, estimate_group_size


def get_gpus_ids(resources_cfg):
    gpus_ids = resources_cfg["gpu_id"]
    if gpus_ids != []:
        return gpus_ids
    # TODO (?) heuristic to pick the best gpus
    return list(range(resources_cfg["gpus"]))


def variable_idxlen_sort(fname):
    return int(fname.split("_")[-1].split(".")[0])


class ChunkedReconstructor:
    """
    A class for doing full-volume reconstructions.
    """

    _pipeline_cls = CudaChunkedPipeline
    _process_name = "reconstruction"

    def __init__(self, process_config, logger=None, extra_options=None, cuda_options=None):
        """
        Initialize a LocalReconstruction object.
        This class is used for managing pipelines

        Parameters
        ----------
        process_config: ProcessConfig object
            Data structure with process configuration
        logger: Logger, optional
            logging object
        extra_options: dict, optional
            Dictionary with advanced options. Please see 'Other parameters' below
        cuda_options: dict, optional
            Dictionary with cuda options passed to `nabu.cuda.processing.CudaProcessing`


        Other parameters
        -----------------
        Advanced options can be passed in the 'extra_options' dictionary. These can be:

           - "gpu_mem_fraction": 0.9,
           - "cpu_mem_fraction": 0.9,
           - "use_phase_margin": True,
           - "max_chunk_size": None,
           - "phase_margin": None,
        """
        self.logger = LoggerOrPrint(logger)
        self.process_config = process_config
        self._set_extra_options(extra_options)
        self._get_reconstruction_range()
        self._get_resources()
        self._compute_max_chunk_size()
        self._compute_phase_margin()
        self._get_pipeline_class()
        self._build_tasks()
        self.pipeline = None
        self.cuda_options = cuda_options


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "gpu_mem_fraction": 0.9,
            "cpu_mem_fraction": 0.9,
            "use_phase_margin": True,
            "max_chunk_size": None,
            "phase_margin": None,
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options
        self.gpu_mem_fraction = self.extra_options["gpu_mem_fraction"]
        self.cpu_mem_fraction = self.extra_options["cpu_mem_fraction"]
        self.use_phase_margin = self.extra_options["use_phase_margin"]
        self._do_histograms = self.process_config.nabu_config["postproc"]["output_histogram"]
        self._histogram_merged = False


    def _get_reconstruction_range(self):
        rec_cfg = self.process_config.nabu_config["reconstruction"]
        self.z_min = rec_cfg["start_z"]
        self.z_max = rec_cfg["end_z"] + 1
        self.delta_z = self.z_max - self.z_min


    def _get_resources(self):
        self.resources = {}
        self._get_gpu()
        self._get_memory()


    def _get_memory(self):
        vm = virtual_memory()
        self.resources["mem_avail_GB"] = vm.available / 1e9


    def _get_gpu(self):
        gpus = get_gpus_ids(self.process_config.nabu_config["resources"])
        if len(gpus) == 0:
            raise ValueError("Need at least one GPU")
        if len(gpus) > 1:
            raise ValueError("This class does not support more than one GPU")
        self.resources["gpu_id"] = self._gpu_id = gpus[0]
        self.resources["gpus"] = collect_cuda_gpus()


    def _compute_max_chunk_size(self):
        gpu_mem = self.resources["gpus"][self._gpu_id]["memory_GB"] * self.gpu_mem_fraction
        cpu_mem = self.resources["mem_avail_GB"] * self.cpu_mem_fraction
        self.gpu_max_chunk_size = estimate_chunk_size(
            gpu_mem,
            self.process_config,
            chunk_step=10,
             warn_from_GB=17 #2**32 elements - see estimate_required_memory docstring note
        )
        self.cpu_max_chunk_size = estimate_chunk_size(
            cpu_mem,
            self.process_config,
            chunk_step=10
        )
        user_max_chunk_size = self.extra_options["max_chunk_size"]
        if user_max_chunk_size is not None:
            self.gpu_max_chunk_size = min(self.gpu_max_chunk_size, user_max_chunk_size)
            self.cpu_max_chunk_size = min(self.cpu_max_chunk_size, user_max_chunk_size)


    def _compute_phase_margin(self):
        unsharp_margin = self._compute_unsharp_margin()
        if "phase" not in self.process_config.processing_steps:
            self._phase_margin = unsharp_margin
            self._margin_v = self._phase_margin[0]
            return
        radio_shape = self.process_config.dataset_info.radio_dims[::-1]
        opts = self.process_config.processing_options["phase"]
        user_phase_margin = self.extra_options["phase_margin"]
        if user_phase_margin is not None and user_phase_margin > 0:
            margin_v, margin_h = user_phase_margin, user_phase_margin
            self.logger.info("Using user-defined phase margin: %d" % user_phase_margin)
        else:
            margin_v, margin_h = compute_paganin_margin(
                radio_shape,
                distance=opts["distance_m"],
                energy=opts["energy_kev"],
                delta_beta=opts["delta_beta"],
                pixel_size=opts["pixel_size_m"],
                padding=opts["padding_type"]
            )
        margin_v = max(margin_v, unsharp_margin[0])
        self._phase_margin = (margin_v, margin_h)
        self._margin_v = self._phase_margin[0]
        self.logger.info("Estimated phase margin: %d pixels" % self._margin_v)


    def _compute_unsharp_margin(self):
        if "unsharp_mask" not in self.process_config.processing_steps:
            return (0, 0)
        opts = self.process_config.processing_options["unsharp_mask"]
        sigma = opts["unsharp_sigma"]
        # nabu uses cutoff = 4
        cutoff = 4
        gaussian_kernel_size = int(ceil(2 * cutoff * sigma + 1))
        self.logger.debug("Unsharp mask margin: %d pixels" % gaussian_kernel_size)
        return (gaussian_kernel_size, gaussian_kernel_size)


    def _get_pipeline_class(self):
        self._limited_mem = False
        # Actually less in some cases (margin_far_up + margin_far_down instead of 2*margin_v).
        # But we want to use only one class for all stages.
        chunk_size_for_one_slice = 1 + 2 * self._margin_v
        chunk_is_too_small = False
        if chunk_size_for_one_slice > self.gpu_max_chunk_size:
            msg = str(
                "Phase margin is %d, so we need to process at least %d detector rows. However, the available memory enables to process only %d rows at once"
                % (self._margin_v, chunk_size_for_one_slice, self.gpu_max_chunk_size)
            )
            chunk_is_too_small = True
        if self._margin_v > self.gpu_max_chunk_size//3:
            n_slices = max(1, self.gpu_max_chunk_size - (2 * self._margin_v))
            n_stages = ceil(self.delta_z / n_slices)
            if n_stages > 1:
                # In this case, using CudaFlatField + margin would lead to many stages
                msg = str(
                    "Phase margin (%d) is too big for chunk size (%d)"
                    % (self._margin_v, self.gpu_max_chunk_size)
                )
                chunk_is_too_small = True
        if chunk_is_too_small:
            self.logger.warning(msg)
            if self.use_phase_margin:
                self._pipeline_cls = CudaChunkedPipelineLimitedMemory
                self.logger.warning("Using %s" % self._pipeline_cls.__name__)
                self._limited_mem = True
            else:
                self._phase_margin = (0, 0)
                self._margin_v = self._phase_margin[0]
                self._pipeline_cls = CudaChunkedPipeline
                self.logger.warning("Using %s without margin" % self._pipeline_cls.__name__)


    def _build_tasks(self):
        self._compute_volume_chunks()


    def _compute_volume_chunks(self):
        n_z = self.process_config.dataset_info._radio_dims_notbinned[1]
        margin_v = self._margin_v
        self._margin_far_up = min(margin_v, self.z_min)
        self._margin_far_down = min(margin_v, n_z - self.z_max)
        # | margin_up |     n_slices    |  margin_down |
        # |-----------|-----------------|--------------|
        # |----------------------------------------------------|
        #                    delta_z
        if self._limited_mem:
            n_slices = self.cpu_max_chunk_size
        else:
            n_slices = self.gpu_max_chunk_size - (2 * margin_v)
        tasks = []
        if self.gpu_max_chunk_size >= self.delta_z and self.z_min == 0 and self.z_max == n_z:
            # In this case we can do everything in a single stage, without margin
            n_slices = self.delta_z
            tasks.append({
                "sub_region": (self.z_min, self.z_max),
                "phase_margin": None,
            })
        elif n_slices >= self.delta_z:
            # In this case we can do everything in a single stage
            n_slices = self.delta_z
            (margin_up, margin_down) = (self._margin_far_up, self._margin_far_down)
            tasks.append({
                "sub_region": (self.z_min - margin_up, self.z_max + margin_down),
                "phase_margin": ((margin_up, margin_down), (0, 0))
            })
        else:
            # In this case there are at least two stages
            n_stages = ceil(self.delta_z / n_slices)
            tasks = []
            curr_z_min = self.z_min
            curr_z_max = self.z_min + n_slices
            for i in range(n_stages):
                margin_up = min(margin_v, curr_z_min)
                margin_down = min(margin_v, max(n_z - curr_z_max, 0))
                if curr_z_max + margin_down >= self.z_max:
                    curr_z_max -= (curr_z_max - (self.z_max + 0))
                    margin_down = min(margin_v, max(n_z - 1 - curr_z_max, 0))
                tasks.append({
                    "sub_region": (curr_z_min - margin_up, curr_z_max + margin_down),
                    "phase_margin": ((margin_up, margin_down), (0, 0))
                })
                if curr_z_max == self.z_max:
                    # No need for further tasks
                    break
                curr_z_min += n_slices
                curr_z_max += n_slices
        self.tasks = tasks
        self.n_slices = n_slices


    def _print_tasks(self):
        for task in self.tasks:
            margin_up, margin_down = task["phase_margin"][0]
            s_u, s_d = task["sub_region"]
            print(
                "Top Margin: [%04d, %04d[  |  Slices: [%04d, %04d[  |  Bottom Margin: [%04d, %04d["
                % (
                    s_u, s_u + margin_up,
                    s_u + margin_up, s_d - margin_down,
                    s_d - margin_down, s_d
                )
            )


    def _instantiate_pipeline(self, task):
        self.logger.debug("Creating a new pipeline object")
        args = [self.process_config, task["sub_region"]]
        kwargs = {}
        if self._pipeline_cls.backend == "cuda":
            kwargs["cuda_options"] = self.cuda_options
        if self._limited_mem:
            # Adapt chunk size so that [margin_up, chunk_size, margin_down]
            # is equal to delta_z.
            chunk_size = self.gpu_max_chunk_size
            dz = self._get_delta_z(task)
            margin_v_tot = sum(task["phase_margin"][0])
            args.append(chunk_size)
        pipeline = self._pipeline_cls(
            *args,
            logger=self.logger,
            phase_margin=task["phase_margin"],
            **kwargs
        )
        self.pipeline = pipeline


    def _instantiate_pipeline_if_necessary(self, current_task, other_task):
        """
        Instantiate a pipeline only if current_task has a different "delta z" than other_task
        """
        if self.pipeline is None:
            self._instantiate_pipeline(current_task)
            return
        dz_cur = self._get_delta_z(current_task)
        dz_other = self._get_delta_z(other_task)
        if dz_cur != dz_other:
            self.logger.debug("Destroying pipeline instance and releasing memory")
            self._destroy_pipeline()
            self._instantiate_pipeline(current_task)


    def _destroy_pipeline(self):
        self.pipeline = None
        # Not elegant, but for now the only way to release Cuda memory
        gc.collect()

    @staticmethod
    def _get_delta_z(task):
        # will have to be modified if sub_region accounts for x-subregion
        return task["sub_region"][1] - task["sub_region"][0]


    def _get_task_key(self):
        """
        Get the 'key' (number) associated to the current task/pipeline
        """
        return self.pipeline.sub_region[-2:]


    def _give_progress_info(self, task):
        self.logger.info("Processing sub-volume %s" % (str(task["sub_region"])))


    def _process_task(self, task):
        self.pipeline.process_chunk(sub_region=task["sub_region"])


    def reconstruct(self):
        tasks = self.tasks
        self.results = {}
        self._histograms = {}
        self._data_dumps = {}
        prev_task = tasks[0]
        for task in tasks:
            self._give_progress_info(task)
            self._instantiate_pipeline_if_necessary(task, prev_task)
            self._process_task(task)
            task_key = self._get_task_key()
            self.results[task_key] = self.pipeline.writer.fname
            if self.pipeline.histogram_writer is not None: # self._do_histograms
                self._histograms[task_key] = self.pipeline.histogram_writer.fname
            if len(self.pipeline._data_dump) > 0:
                self._data_dumps[task_key] = {}
                for step_name, writer in self.pipeline._data_dump.items():
                    self._data_dumps[task_key][step_name] = writer.fname
            prev_task = task


    def get_relative_files(self, files=None):
        out_cfg = self.process_config.nabu_config["output"]
        if files is None:
            files = list(self.results.values())
        try:
            files.sort(key=variable_idxlen_sort)
        except:
            self.logger.error(
                "Lexical ordering failed, falling back to default sort - it will fail for more than 10k projections"
            )
            files.sort()
        local_files = [
            join(out_cfg["file_prefix"], basename(fname))
            for fname in files
        ]
        return local_files


    def merge_hdf5_reconstructions(self, output_file=None, prefix=None, files=None, process_name=None, axis=0, merge_histograms=True, output_dir=None):
        """
        Merge existing hdf5 files by creating a HDF5 virtual dataset.

        Parameters
        ----------
        output_file: str, optional
            Output file name. If not given, the file prefix in section "output"
            of nabu config will be taken.
        """
        out_cfg = self.process_config.nabu_config["output"]
        out_dir = output_dir or out_cfg["location"]
        prefix = prefix or ""
        # Prevent issue when out_dir is empty, which happens only if dataset/location is a relative path.
        # TODO this should be prevented earlier
        if out_dir is None or len(out_dir.strip()) == 0:
            out_dir = dirname(dirname(self.results[list(self.results.keys())[0]]))
        #
        if output_file is None:
            output_file = join(out_dir, prefix + out_cfg["file_prefix"]) + ".hdf5"
        if isfile(output_file):
            msg = str("File %s already exists" % output_file)
            if out_cfg["overwrite_results"]:
                msg += ". Overwriting as requested in configuration file"
                self.logger.warning(msg)
            else:
                msg += ". Set overwrite_results to True in [output] to overwrite existing files."
                self.logger.fatal(msg)
                raise ValueError(msg)

        local_files = files
        if local_files is None:
            local_files = self.get_relative_files()
        if local_files == []:
            self.logger.error("No files to merge")
            return
        entry = getattr(self.process_config.dataset_info.dataset_scanner, "entry", "entry")
        process_name = process_name or self._process_name
        h5_path = join(entry, *[process_name, "results", "data"])
        #
        self.logger.info("Merging %ss to %s" % (process_name, output_file))
        merge_hdf5_files(
            local_files, h5_path, output_file, process_name,
            output_entry=entry,
            output_filemode="a",
            processing_index=0,
            config={
                self._process_name + "_stages": {
                    str(k): v for k, v in zip(self.results.keys(), local_files)
                },
                "nabu_config": self.process_config.nabu_config,
                "processing_options": self.process_config.processing_options,
            },
            base_dir=out_dir,
            axis=axis,
            overwrite=out_cfg["overwrite_results"]
        )
        if merge_histograms:
            self.merge_histograms(output_file=output_file)
        return output_file

    merge_hdf5_files = merge_hdf5_reconstructions

    def merge_histograms(self, output_file=None, force_merge=False):
        """
        Merge the partial histograms
        """
        if not(self._do_histograms):
            return
        if self._histogram_merged and not(force_merge):
            return
        self.logger.info("Merging histograms")

        masterfile_entry = getattr(self.process_config.dataset_info.dataset_scanner, "entry", "entry")
        masterfile_process_name = "histogram" # TODO don't hardcode process name
        output_entry = masterfile_entry

        out_cfg = self.process_config.nabu_config["output"]
        if output_file is None:
            output_file = join(
                dirname(list(self._histograms.values())[0]),
                out_cfg["file_prefix"] + "_histogram"
            ) + ".hdf5"
        local_files = self.get_relative_files(files=list(self._histograms.values()))
        #
        h5_path = join(masterfile_entry, *[masterfile_process_name, "results", "data"])
        #

        try:
            files = sorted(self._histograms.values(), key=variable_idxlen_sort)
        except:
            self.logger.error(
                "Lexical ordering of histogram failed, falling back to default sort - it will fail for more than 10k projections"
            )
            files = sorted(self._histograms.values())
        data_urls = []
        for fname in files:
            url = DataUrl(
                file_path=fname, data_path=h5_path, data_slice=None, scheme="silx"
            )
            data_urls.append(url)
        histograms = []
        for data_url in data_urls:
            h2D = get_data(data_url)
            histograms.append(
                (h2D[0], add_last_bin(h2D[1]))
            )
        histograms_merger = PartialHistogram( # TODO configurable
            method="fixed_bins_number", num_bins=histograms[0][0].size
        )
        merged_hist = histograms_merger.merge_histograms(histograms)

        #volume_shape = (self.delta_z, ) + self.process_config.dataset_info.radio_dims[::-1]
        rec_options = self.process_config.processing_options["reconstruction"]
        volume_shape = (
            rec_options["end_z"] - rec_options["start_z"] + 1,
            rec_options["end_y"] - rec_options["start_y"] + 1,
            rec_options["end_x"] - rec_options["start_x"] + 1,
        )
        writer = NXProcessWriter(
            output_file, entry=output_entry, filemode="a", overwrite=True
        )
        writer.write(
            hist_as_2Darray(merged_hist),
            "histogram", # TODO don't hard-code
            processing_index=1,
            config={
                "files": local_files,
                "bins": self.process_config.nabu_config["postproc"]["histogram_bins"],
                "volume_shape": volume_shape,
            }
        )
        self._histogram_merged = True


    def merge_data_dumps(self, axis=1):
        # Collect in a dict where keys are step names (instead of task keys)
        dumps = {}
        for task_key, data_dumps in self._data_dumps.items():
            for step_name, fname in data_dumps.items():
                fname = join(basename(dirname(fname)), basename(fname))
                if step_name not in dumps:
                    dumps[step_name] = [fname]
                else:
                    dumps[step_name].append(fname)
        # Merge HDF5 files
        for step_name, files in dumps.items():
            dump_file = self.process_config.get_save_steps_file(step_name=step_name)
            self.merge_hdf5_files(
                output_file=dump_file,
                output_dir=dirname(dump_file),
                files=files, process_name=step_name, axis=axis, merge_histograms=False
            )


# COMPAT.
LocalReconstruction = ChunkedReconstructor
#

# TODO this class inherits from ChunkedReconstructor, but its principle is different.
# Many methods from ChunkedReconstructor are reused.
# Create a base class ?
class GroupedReconstructor(ChunkedReconstructor):

    _pipeline_cls_radios = CudaGroupedPipeline
    _pipeline_cls_sinos = CudaSinoStackPipeline
    _reconstruction_states = ["radios", "sinos"]

    def __init__(self, process_config, logger=None, extra_options=None, cuda_options=None):
        self.logger = LoggerOrPrint(logger)
        self.process_config = process_config
        self._set_extra_options(extra_options)
        self._get_reconstruction_range()
        self._get_subregion()
        self._get_resources()
        self._compute_max_group_size() # radios pipeline
        self._compute_max_chunk_size() # sinos pipeline
        self._build_tasks()
        self.pipeline = None
        self.set_reconstruction_state("radios")
        self.cuda_options = cuda_options


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "gpu_mem_fraction": 0.9,
            "cpu_mem_fraction": 0.9,
            "max_group_size": None,
            "max_chunk_size": None,
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options
        self.gpu_mem_fraction = self.extra_options["gpu_mem_fraction"]
        self.cpu_mem_fraction = self.extra_options["cpu_mem_fraction"]
        self._histogram_merged = False


    def _get_subregion(self):
        # For GroupedPipeline, sub-region is always the same
        x_min, x_max = None, None

        # TODO the following is not working for half tomography, as end_x is "twice" bigger.
        # So X-subregion is not supported until the "config file ingestion refactoring"
        # if "reconstruction" in self.process_config.processing_steps:
            # rec_cfg = self.process_config.processing_options["reconstruction"]
            # x_min, x_max = rec_cfg["start_x"], rec_cfg["end_x"] + 1
        self.sub_region = (x_min, x_max, self.z_min, self.z_max)

    def _compute_max_group_size(self):
        """
        Compute the maximum number of radios that can be processed in memory
        """
        gpu_mem = self.resources["gpus"][self._gpu_id]["memory_GB"] * self.gpu_mem_fraction
        cpu_mem = self.resources["mem_avail_GB"] * self.cpu_mem_fraction
        self.gpu_max_group_size = estimate_group_size(
            gpu_mem,
            self.process_config,
            step=10,
            warn_from_GB=17 #2**32 elements - see estimate_required_memory docstring note
        )
        self.cpu_max_group_size = estimate_group_size(
            cpu_mem,
            self.process_config,
            step=10,
        )
        user_max_group_size = self.extra_options["max_group_size"]
        if user_max_group_size is not None:
            self.gpu_max_group_size = min(self.gpu_max_group_size, user_max_group_size)
            self.cpu_max_group_size = min(self.cpu_max_group_size, user_max_group_size)

        self.max_group_size = self.gpu_max_group_size

    def _compute_max_chunk_size(self):
        processing_steps = self.process_config.processing_steps
        processing_steps, removed_steps = remove_items_from_list(
            processing_steps,
            ["flatfield", "ccd_correction", "phase"]
        )
        LocalReconstruction._compute_max_chunk_size(self)
        restore_items_in_list(processing_steps, removed_steps)


    def _build_tasks(self):
        self._build_radios_processing_tasks()
        self._build_sinos_processing_tasks()

    def _build_radios_processing_tasks(self):
        n_a = len(self.process_config.dataset_info.projections)
        group_size = self.max_group_size
        tasks = []
        if group_size >= n_a:
            # In this case we can do everything in a single stage
            # (but then we would have been better of with a LocalReconstruction)
            tasks.append({
                "images_group": (0, n_a),
            })
        else:
            # In this case there are at least two stages
            n_stages = ceil(n_a / group_size)
            tasks = []
            curr_a_min = 0
            curr_a_max = group_size
            for i in range(n_stages):
                curr_a_max = min(n_a, curr_a_max)
                tasks.append({
                    "images_group": (curr_a_min, curr_a_max)
                })
                if curr_a_max == n_a:
                    # No need for further tasks
                    break
                curr_a_min += group_size
                curr_a_max += group_size
        self.radios_tasks = tasks


    def _build_sinos_processing_tasks(self):
        n_z = self.delta_z
        chunk_size = self.gpu_max_chunk_size
        tasks = []
        if chunk_size >= n_z:
            # In this case we can do everything in a single stage
            # (but then we would have been better of with a LocalReconstruction)
            tasks.append({
                "stack": (0, n_z),
            })
        else:
            # In this case there are at least two stages
            n_stages = ceil(n_z / chunk_size)
            tasks = []
            curr_a_min = 0
            curr_a_max = chunk_size
            for i in range(n_stages):
                curr_a_max = min(n_z, curr_a_max)
                tasks.append({
                    "stack": (curr_a_min, curr_a_max)
                })
                if curr_a_max == n_z:
                    # No need for further tasks
                    break
                curr_a_min += chunk_size
                curr_a_max += chunk_size
        self.sinos_tasks = tasks


    def set_reconstruction_state(self, state):
        """
        Set the reconstruction state of the current class, depending on which stage of processing
        we want to accomplish.

        Parameters
        ----------
        state: str
            Reconstruction state. Can be 'radios' or 'sinos'
        """
        if state == "radios":
            self._process_name = "sinogram"
            self._do_histograms = False
            self.tasks = self.radios_tasks
        elif state == "sinos":
            self._process_name = "reconstruction"
            self._do_histograms = self.process_config.nabu_config["postproc"]["output_histogram"]
            self.tasks = self.sinos_tasks
        else:
            raise ValueError("state can only be in %s" % str(self._reconstruction_states))
        self._reconstruction_state = state


    def _print_tasks(self):
        if self._reconstruction_state == "radios":
            desc = "Images"
            key = "images_group"
        else:
            desc = "Sinogram stack"
            key = "stack"
        for task in self.tasks:
            i_u, i_d = task[key]
            print(
                "%s: [%04d, %04d[" % desc
                % (i_u, i_d)
            )

    @staticmethod
    def _get_group_size(task):
        group = task["images_group"]
        return group[1] - group[0]

    @staticmethod
    def _get_stack_size(task):
        stack = task["stack"]
        return stack[1] - stack[0]


    def _instantiate_pipeline_radios(self, task):
        args = [self.process_config, task["images_group"]]
        kwargs = {}
        if self._pipeline_cls_radios.backend == "cuda":
            kwargs["cuda_options"] = self.cuda_options
        pipeline = self._pipeline_cls_radios(
            *args,
            sub_region=self.sub_region,
            logger=self.logger,
            **kwargs
        )
        self.pipeline = pipeline


    def _instantiate_pipeline_sinos(self, task):
        args = [self.process_config, task["stack"]]
        kwargs = {}
        if self._pipeline_cls_sinos.backend == "cuda":
            kwargs["cuda_options"] = self.cuda_options
        pipeline = self._pipeline_cls_sinos(
            *args,
            self._projections,
            logger=self.logger,
            **kwargs
        )
        self.pipeline = pipeline


    def _instantiate_pipeline(self, task):
        self.logger.debug("Creating a new pipeline object")
        if self._reconstruction_state == "radios":
            self._instantiate_pipeline_radios(task)
        else:
            self._instantiate_pipeline_sinos(task)


    def _instantiate_pipeline_if_necessary(self, current_task, other_task):
        """
        Instantiate a pipeline only if current_task has a different "group_size" than other_task
        """
        if self.pipeline is None:
            self._instantiate_pipeline(current_task)
            return
        if self._reconstruction_state == "radios":
            group_size_cur = self._get_group_size(current_task)
            group_size_other = self._get_group_size(other_task)
            recreate_pipeline = group_size_cur != group_size_other
        else:
            stack_size_cur = self._get_stack_size(current_task)
            stack_size_other = self._get_stack_size(other_task)
            recreate_pipeline = stack_size_cur != stack_size_other
        if recreate_pipeline:
            self.logger.debug("Destroying pipeline instance and releasing memory")
            self._destroy_pipeline()
            self._instantiate_pipeline(current_task)


    def _get_task_key(self):
        """
        Get the 'key' (number) associated to the current task/pipeline
        """
        if self._reconstruction_state == "radios":
            return self.pipeline.images_group
        else:
            return self.pipeline.stack


    def _give_progress_info(self, task):
        if self._reconstruction_state == "radios":
            self.logger.info("Processing images group %s" % (str(task["images_group"])))
        else:
            self.logger.info("Processing sinograms stack %s" % (str(task["stack"])))


    def _process_task(self, task):
        if self._reconstruction_state == "radios":
            self.pipeline.process_group(images_group=task["images_group"])
        else:
            self.pipeline.process_stack(stack=task["stack"])


    def reconstruct(self, projections=None):
        """
        Perform a volume reconstruction with the current configuration.
        """
        if self.process_config.resume_from_step is not None and self.process_config.resume_from_step == "sinogram":
            projections = self.process_config.processing_options["read_chunk"]["process_file"]
            self.logger.info("Resuming from sinograms in %s" % projections)
        if projections is None:
            self.set_reconstruction_state("radios")
            # TODO this is clumsy - see comment above the class definition
            LocalReconstruction.reconstruct(self)
            #
            self._projections = self.merge_hdf5_reconstructions(prefix="sinogram_", merge_histograms=False)
            self.merge_data_dumps(axis=0)
            self._destroy_pipeline()
        else:
            self._projections = projections
        self.set_reconstruction_state("sinos")
        LocalReconstruction.reconstruct(self)



from ..reconstruction.sinogram_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.preproc.sinogram_cuda was moved to nabu.reconstruction.sinogram_cuda.",
    func_name="preproc_sinogram"
)


from os import path
from tempfile import mkdtemp
import pytest
import numpy as np
from silx.io.url import DataUrl
from nabu.testutils import NXDatasetMock
from nabu.resources.nxflatfield import NXFlatField, val_to_nxkey


tempdir = mkdtemp(prefix="nabu_")


test_nxflatfield_scenarios = [
    {
        "name": "simple",
        "flats_pos": [slice(1, 6)],
        "darks_pos": [slice(0, 1)],
        "flats_reduction": "mean",
        "darks_reduction": "mean",
        "results_file": None
    },
    {
        "name": "simple_with_save",
        "flats_pos": [slice(1, 6)],
        "darks_pos": [slice(0, 1)],
        "flats_reduction": "mean",
        "darks_reduction": "mean",
        "results_file": path.join(tempdir, "ff_simple_with_save.h5"),
    },
    {
        "name": "multiple_with_save",
        "flats_pos": [slice(0, 10), slice(30, 40)],
        "darks_pos": [slice(95, 100)],
        "flats_reduction": "median",
        "darks_reduction": "mean",
        "results_file": path.join(tempdir, "ff_multiple_with_save.h5"),
    },

]

def generate_nx_dataset(fname, n_projs, flats_pos, darks_pos, h5path=None):
    image_keys = np.zeros(n_projs, dtype=np.int64)
    data_volume = np.random.randint(0, high=65535, size=(n_projs, 3, 4)).astype(np.uint16)
    # Flats have values 1, 2, 3, ...
    for i, slice_ in enumerate(flats_pos):
        for j in range(slice_.start, slice_.stop):
            data_volume[j] = j + i + 1
            image_keys[j] = val_to_nxkey["flats"]
    # Darks have values 1000, 1001, 1002, ...
    for i, slice_ in enumerate(darks_pos):
        for j in range(slice_.start, slice_.stop):
            data_volume[j] = j + i + 1000
            image_keys[j] = val_to_nxkey["darks"]
    nx_dataset_mock = NXDatasetMock(
        data_volume, image_keys
    )
    nx_dataset_mock.generate_hdf5_file(fname, h5path=h5path)
    return nx_dataset_mock


# parametrize with fixture and "params=" will launch a new class for each scenario.
# the attributes set to "cls" will remain for all the tests done in this class
# with the current scenario.
@pytest.fixture(scope='class', params=test_nxflatfield_scenarios)
def bootstrap(request):
    cls = request.cls
    cls.n_projs = 100
    cls.params = request.param


@pytest.mark.usefixtures('bootstrap')
class TestNXFlatField:

    def get_h5_filename(self):
        return path.join(tempdir, "dataset_" + self.params["name"] + ".h5")

    def _get_expected_results_imgtype(self, data_volume, what):
        slices_pos = self.params[what + "_pos"]
        reduction_func = getattr(np, self.params[what + "_reduction"])
        ref = {}
        for slice_ in slices_pos:
            ref[slice_.start] = reduction_func(
                data_volume[slice_], axis=0
            )
        return ref

    def get_expected_results(self, dataset_mock):
        data_volume = dataset_mock.dataset_dict["instrument"]["detector"]["data"]
        ref = {}
        ref["flats"] = self._get_expected_results_imgtype(data_volume, "flats")
        ref["darks"] = self._get_expected_results_imgtype(data_volume, "darks")
        return ref

    def compare_results(self, res, ref):
        assert res.keys() == ref.keys()
        for what in ref.keys():
            assert res[what].keys() == ref[what].keys()
            for idx, arr in res[what].items():
                assert np.allclose(arr, ref[what][idx])


    def test_nxflatfield(self):
        dataset_fname = self.get_h5_filename()
        nx_dataset = generate_nx_dataset(
            dataset_fname,
            self.n_projs,
            self.params["flats_pos"],
            self.params["darks_pos"],
            h5path="/entry"
        )

        results_url = None
        if self.params["results_file"] is not None:
            results_url = DataUrl(
                file_path=self.params["results_file"],
                data_path="/entry0000/flat_field_images"
            )
        nxff = NXFlatField(
            nx_dataset,
            results_url=results_url,
            flats_reduction=self.params["flats_reduction"],
            darks_reduction=self.params["darks_reduction"]
        )
        if results_url is not None:
            res_url = nxff.get_final_urls()
            res = nxff.load_data(res_url)
        else:
            res = nxff.compute_final_images()
        ref = self.get_expected_results(nx_dataset)
        self.compare_results(res, ref)
        if results_url is not None:
            nxff2 = NXFlatField(
                nx_dataset,
                lookup_files=[results_url],
                flats_reduction=self.params["flats_reduction"],
                darks_reduction=self.params["darks_reduction"]
            )
            res2_url = nxff2.get_final_urls()
            res2 = nxff.load_data(res2_url)
            self.compare_results(res2, ref)




#!/usr/bin/env python

from .. import version
from ..resources.utils import is_hdf5_extension
from ..pipeline.config import parse_nabu_config_file
from ..pipeline.config_validators import convert_to_int
from .cli_configs import ReconstructConfig
from .utils import parse_params_values


def update_reconstruction_start_end(conf_dict, user_indices):
    if len(user_indices) == 0:
        return
    rec_cfg = conf_dict["reconstruction"]
    err = None

    val_int, conv_err = convert_to_int(user_indices)
    if conv_err is None:
        start_z = user_indices
        end_z = user_indices
    else:
        if user_indices in ["first", "middle", "last"]:
            start_z = user_indices
            end_z = user_indices
        elif user_indices == "all":
            start_z = 0
            end_z = -1
        elif "-" in user_indices:
            try:
                start_z, end_z = user_indices.split("-")
                start_z = int(start_z)
                end_z = int(end_z)
            except Exception as exc:
                err = "Could not interpret slice indices '%s': %s" % (user_indices, str(exc))
        else:
            err = "Could not interpret slice indices: %s" % user_indices
        if err is not None:
            print(err)
            exit(1)
    rec_cfg["start_z"] = start_z
    rec_cfg["end_z"] = end_z


def get_log_file(arg_logfile, legacy_arg_logfile, forbidden=None):
    default_arg_val = ""
    # Compat. log_file --> logfile
    if legacy_arg_logfile != default_arg_val:
        logfile = legacy_arg_logfile
    else:
        logfile = arg_logfile
    #
    if forbidden is None:
        forbidden = []
    for forbidden_val in forbidden:
        if logfile == forbidden_val:
            print("Error: --logfile argument cannot have the value %s" % forbidden_val)
            exit(1)
    if logfile == "":
        logfile = True
    return logfile


def main():
    args = parse_params_values(
        ReconstructConfig, parser_description="Perform a tomographic reconstruction.", program_version="nabu " + version
    )

    # Imports are done here, otherwise "nabu --version" takes forever
    from ..pipeline.fullfield.processconfig import ProcessConfig
    from ..cuda.utils import __has_pycuda__

    if __has_pycuda__:
        from ..pipeline.fullfield.local_reconstruction import ChunkedReconstructor, GroupedReconstructor
    else:
        print("Error: need cuda and pycuda for reconstruction")
        exit(1)
    from ..resources.logger import Logger

    #

    # A crash with scikit-cuda happens only on PPC64 platform if and nvidia-persistenced is running.
    # On such machines, a warm-up has to be done.
    import platform

    if platform.machine() == "ppc64le":
        from silx.math.fft.cufft import CUFFT
    #

    logfile = get_log_file(args["logfile"], args["log_file"], forbidden=[args["input_file"]])
    conf_dict = parse_nabu_config_file(args["input_file"])
    update_reconstruction_start_end(conf_dict, args["slice"].strip())

    proc = ProcessConfig(conf_dict=conf_dict, create_logger=logfile)
    logger = proc.logger

    logger.info(
        "Going to reconstruct slices (%d, %d)"
        % (proc.nabu_config["reconstruction"]["start_z"], proc.nabu_config["reconstruction"]["end_z"])
    )

    # (hopefully) temporary patch
    if "phase" in proc.processing_steps:
        if args["energy"] > 0:
            logger.warning("Using user-provided energy %.2f keV" % args["energy"])
            proc.dataset_info.dataset_scanner._energy = args["energy"]
            proc.processing_options["phase"]["energy_kev"] = args["energy"]
        if proc.dataset_info.energy < 1e-3 and proc.nabu_config["phase"]["method"] is not None:
            msg = "No information on energy. Cannot retrieve phase. Please use the --energy option"
            logger.fatal(msg)
            raise ValueError(msg)
    #

    # Determine which reconstructor to use
    reconstructor_cls = ChunkedReconstructor
    phase_method = None
    if "phase" in proc.processing_steps:
        phase_method = proc.processing_options["phase"]["method"]

    # Determine which pipeline should be used:
    #  - CTF filter needs to process full radios
    #  - Projections rotation can be chunked only for small angles
    if phase_method == "CTF":
        reconstructor_cls = GroupedReconstructor
    if ("rotate_projections" in proc.processing_steps) and abs(proc.processing_options["rotate_projections"]["angle"]) > 15:
        reconstructor_cls = GroupedReconstructor
    if bool(args["force_use_grouped_pipeline"]):
        reconstructor_cls = GroupedReconstructor
    logger.debug("Using pipeline: %s" % reconstructor_cls.__name__)


    # Get extra options
    extra_options = {
        "gpu_mem_fraction": args["gpu_mem_fraction"],
        "cpu_mem_fraction": args["cpu_mem_fraction"],
    }
    if reconstructor_cls is ChunkedReconstructor:
        extra_options.update(
            {
                "use_phase_margin": args["use_phase_margin"],
                "max_chunk_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
                "phase_margin": args["phase_margin"],
            }
        )
    else:
        extra_options.update(
            {
                "max_group_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
            }
        )

    R = reconstructor_cls(proc, logger=logger, extra_options=extra_options)

    R.reconstruct()
    R.merge_data_dumps()
    if is_hdf5_extension(proc.nabu_config["output"]["file_format"]):
        R.merge_hdf5_reconstructions()
    R.merge_histograms()


if __name__ == "__main__":
    main()



import numpy as np
import pycuda.gpuarray as garray
from ..cuda.kernel import CudaKernel
from ..utils import get_cuda_srcfile, updiv, deprecated_class
from .sinogram import SinoBuilder, SinoNormalization
from .sinogram import _convert_halftomo_right # FIXME Temporary patch
from ..cuda.processing import CudaProcessing


class CudaSinoBuilder(SinoBuilder):
    def __init__(self, sinos_shape=None, radios_shape=None, rot_center=None, halftomo=False, cuda_options=None):
        """
        Initialize a CudaSinoBuilder instance.
        Please see the documentation of nabu.reconstruction.sinogram.Builder
        and nabu.cuda.processing.CudaProcessing.
        """
        super().__init__(
            sinos_shape=sinos_shape, radios_shape=radios_shape, rot_center=rot_center,
            halftomo=halftomo,
        )
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_cuda_halftomo()


    def _init_cuda_halftomo(self):
        if not(self.halftomo):
            return
        kernel_name = "halftomo_kernel"
        self.halftomo_kernel = CudaKernel(
            kernel_name,
            get_cuda_srcfile("halftomo.cu"),
            signature="PPPiii",
        )
        rc = self._rot_center_int
        blk = (32, 32, 1) # tune ?
        self._halftomo_blksize = blk
        self._halftomo_gridsize = (
            updiv(2 * rc, blk[0]),
            updiv((self.n_angles + 1)//2, blk[1]),
            1
        )
        d = self.n_x - rc # will have to be adapted for varying axis pos
        self.halftomo_weights = np.linspace(0, 1, 2*abs(d), endpoint=True, dtype="f")
        self.d_halftomo_weights = garray.to_gpu(self.halftomo_weights)
        if self._halftomo_flip:
            self.xflip_kernel = CudaKernel(
                "reverse2D_x",
                get_cuda_srcfile("ElementOp.cu"),
                signature="Pii"
            )
            blk = (32, 32, 1)
            self._xflip_blksize = blk
            self._xflip_gridsize_1 = (
                updiv(self.n_x, blk[0]),
                updiv(self.n_angles, blk[1]),
                1
            )
            self._xflip_gridsize_2 = self._halftomo_gridsize


    # Overwrite parent method
    def _radios_to_sinos_simple(self, radios, output, copy=False):
        if not(copy) and output is None:
            return radios.transpose(axes=(1, 0, 2)) # view
        if output is None: # copy and output is None
            na, nz, nx = radios.shape
            output = garray.zeros((nz, na, nx), "f")
        # not(copy) and output is not None
        for i in range(output.shape[0]):
            output[i, :, :] = radios[:, i, :]
        return output


    # Overwrite parent method
    def _radios_to_sinos_halftomo(self, radios, sinos):
        n_a, n_z, n_x = radios.shape
        n_a2 = (n_a + 1) // 2
        rc = self._rot_center_int
        out_dwidth = 2 * rc
        if sinos is not None:
            if sinos.shape[-1] != out_dwidth:
                raise ValueError(
                    "Expected sinos sinogram last dimension to have %d elements"
                    % out_dwidth
                )
            if sinos.shape[-2] != n_a2:
                raise ValueError("Expected sinograms to have %d angles" % n_a2)
        else:
            sinos = garray.zeros(self.sinos_halftomo_shape, dtype=np.float32)

        # FIXME: TEMPORARY PATCH, waiting for cuda implementation
        if self._rot_center_int > self.n_x:
            return self._radios_to_sinos_halftomo_external_cor(radios, sinos)
        #

        # need to use a contiguous 2D, array otherwise kernel does not work
        if n_a & 1:
            d_sino = garray.zeros((n_a + 1, n_x), "f")
        else:
            d_sino = radios[:, 0, :].copy()
        for i in range(n_z):
            d_sino[:n_a] = radios[:, i, :]
            if n_a & 1:
                d_sino[-1, :].fill(0)
            if self._halftomo_flip:
                self.xflip_kernel(
                    d_sino, n_x, n_a,
                    grid=self._xflip_gridsize_1, block=self._xflip_blksize
                )
            self.halftomo_kernel(
                d_sino,
                sinos[i],
                self.d_halftomo_weights,
                n_a, n_x, rc,
                grid=self._halftomo_gridsize,
                block=self._halftomo_blksize
            )
            if self._halftomo_flip:
                self.xflip_kernel(
                    sinos[i], 2*rc, n_a2,
                    grid=self._xflip_gridsize_2, block=self._xflip_blksize
                )
        return sinos


    def _radios_to_sinos_halftomo_external_cor(self, radios, sinos):
        """
        TEMPORARY PATCH waiting to have a cuda implementation
        Processing is done by getting reach radio on host, which is suboptimal
        """
        n_a, n_z, n_x = radios.shape
        n_a2 = n_a // 2
        rc = self._rot_center_int
        out_dwidth = 2 * rc

        # need to use a contiguous 2D, array otherwise kernel does not work
        sino = radios[:, 0, :].get()
        for i in range(n_z):
            radios[:, i, :].get(sino)
            if self._halftomo_flip:
                sino = sino[:, ::-1]
            sino_half = _convert_halftomo_right(sino, self._rot_center_int)
            sinos[i, :, :] = sino_half[:]
            if self._halftomo_flip:
                self.xflip_kernel(
                    sinos[i], 2*rc, n_a2,
                    grid=self._xflip_gridsize_2, block=self._xflip_blksize
                )
        return sinos


CudaSinoProcessing = deprecated_class(
    "'CudaSinoProcessing' was renamed 'CudaSinoBuilder'", do_print=True
)(CudaSinoBuilder)


class CudaSinoNormalization(SinoNormalization):
    def __init__(self, kind="chebyshev", sinos_shape=None, radios_shape=None, cuda_options=None):
        super().__init__(
            kind=kind, sinos_shape=sinos_shape, radios_shape=radios_shape,
        )
        self._get_shapes(sinos_shape, radios_shape)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_cuda_normalization()


    _get_shapes = SinoBuilder._get_shapes

    def _init_cuda_normalization(self):
        self._cuda_kernel = CudaKernel(
            self._get_norm_func_name(),
            filename=get_cuda_srcfile("normalization.cu"),
            signature="Piii",
        )
        self._cuda_kernel_args = [
            np.int32(self.n_x), np.int32(self.n_angles), np.int32(self.n_z)
        ]
        blk = (1, 64, 16) # TODO tune ?
        self._cuda_kernel_kwargs = {
            "block": blk,
            "grid": (1, int(updiv(self.n_angles, blk[1])), int(updiv(self.n_z, blk[2]))),
        }


    def normalize(self, sinos):
        if sinos.flags.c_contiguous:
            self._cuda_kernel(
                sinos, *self._cuda_kernel_args, **self._cuda_kernel_kwargs
            )
        else:
            d_tmp = garray.zeros(sinos.shape[1:], "f")
            # This kernel seems to have an issue on arrays that are not C-contiguous.
            # We have to process image per image.
            nz = np.int32(1)
            nthreadsperblock = (1, 32, 1) # TODO tune
            nblocks = (1, int(updiv(self.n_angles, nthreadsperblock[1])), 1)
            for i in range(sinos.shape[0]):
                d_tmp[:] = sinos[i][:]
                self._cuda_kernel(
                    d_tmp,
                    np.int32(self.n_x), np.int32(self.n_angles), np.int32(1),
                    grid=nblocks,
                    block=nthreadsperblock
                )
                sinos[i][:] = d_tmp[:]
        return sinos



from os import path
import posixpath
import numpy as np
from ...preproc.ctf import CTFPhaseRetrieval, GeoPars
from ...resources.logger import LoggerOrPrint
from ...utils import recursive_copy_dict, ArrayPlaceHolder
from ...io.reader import ChunkReader, HDF5Loader, load_images_stack_from_hdf5, get_hdf5_dataset_shape
from ...io.utils import get_first_hdf5_entry
from ..utils import use_options, pipeline_step, WriterConfigurator
from .chunked import ChunkedPipeline


# TODO for now this class inherits from ChunkedPipeline, although its principle is quite different.
# But many methods can be re-used from ChunkedPipeline.
# Create a base class ?
class GroupedPipeline(ChunkedPipeline):
    """
    A class for processig groups of radios.
    """
    backend = "numpy"

    def __init__(self, process_config, images_group, sub_region=None, logger=None, extra_options=None):
        """
        Initialize a GroupedPipeline object.

        Parameters
        ----------
        process_config: ProcessConfig
            Pipeline description.
        images_group: tuple of two int
            Images group to process, in the form (start, end) where end-start == group_size.
            The "end" index is not included, as in standard Python indexing.
            This parameter can be reset while calling process_group(), in order to process another group of images.
        sub_region: tuple, optional
            sub-region to consider for each image, in the form `(start_x, end_x)` or `(start_x, end_x, start_z, end_z)`.
            If provided, each image will be cropped as `image[:, start_x:end_x]` or `image[start_z:end_z, start_x:end_x]`
        logger: Logger, optional
            logging object. If not provided, a new logging object is created.
        extra_options: dict, optional
            dict of advanced options. See "Other Parameters" for these options.


        Other parameters
        ----------------
        Advanced parameters can be passed through the "extra_options" dict.
        Current options are: None


        Important
        ----------
        This class is to be used when you want to process images by groups, not by chunks.
        Contrarily to `ChunkedPipeline`, the sub_region (start_x, end_x, start_z, end_z) cannot be changed.
        What can be changed is only the "images group".
        In short:

           - Use `ChunkedPipeline` to process `images[:, 0:100, :]` then `images[:, 100:200, :]`, etc
           - Use `GroupedPipeline` to process `images[0:1000, :, :]` then `images[1000:2000, :, :]`, etc

        """
        self.logger = LoggerOrPrint(logger)
        self._set_params(process_config, images_group, sub_region, extra_options)
        self._init_pipeline()


    def _set_params(self, process_config, images_group, sub_region, extra_options):
        self.process_config = process_config
        self._phase_margin = None
        self.dataset_info = self.process_config.dataset_info
        self.dataset_infos = self.process_config.dataset_info # other shorthand - deprecated
        self.processing_steps = self.process_config.processing_steps.copy()
        self._orig_processing_options = recursive_copy_dict(self.process_config.processing_options)
        self.processing_options = recursive_copy_dict(self.process_config.processing_options)

        self._set_group(images_group)
        self._set_subregion(sub_region)
        self._set_extra_options(extra_options)
        self._callbacks = {}
        self._orig_file_prefix = None
        self._steps_name2component = {}
        self._steps_component2name = {}
        self._data_dump = {}
        self._phase_margin_up = 0 # Compat.
        self._resume_from_step = None


    def _set_group(self, images_group):
        self.images_group = images_group
        self.group_size = images_group[1] - images_group[0]


    def _set_subregion(self, sub_region):
        if sub_region is None:
            sub_region = (None, None, None, None)
        if len(sub_region) == 4:
            start_x, end_x, start_z, end_z = sub_region
        elif len(sub_region) == 2:
            first_part, second_part = sub_region
            if np.iterable(first_part): # ((sx, ex), (sz, ez))
                start_x, end_x = first_part
                start_z, end_z = second_part
            else: # (sx, ex, sz, ez)
                start_x, end_x = first_part, second_part
                start_z, end_z = None, None
        self.sub_region = (start_x, end_x, start_z, end_z)
        self.z_min = start_z or 0


    def _get_phase_margin(self):
        return ((0, 0), (0, 0))


    def _get_shape(self, step_name):
        radios_3D_shape = self.radios.shape
        radio_2D_shape = radios_3D_shape[1:]
        shapes = {
            "flatfield": radios_3D_shape,
            "double_flatfield": radios_3D_shape,
            "ccd_correction": radio_2D_shape,
            "phase": radio_2D_shape,
            "unsharp_mask": radio_2D_shape,
            "take_log": radios_3D_shape,
            "radios_movements": radios_3D_shape,
            "rotate_projections": radio_2D_shape,
        }
        if step_name not in shapes:
            raise ValueError("Cannot figure out shape for step '%s'" % step_name)
        shape = shapes[step_name]
        self.logger.debug("Data shape for %s is %s" % (step_name, str(shape)))
        return shape


    def _check_resume_from_step(self):
        if self._resume_from_step is None:
            return
        expected_radios_shape = get_hdf5_dataset_shape(
            self.processing_options["read_chunk"]["process_file"],
            self.processing_options["read_chunk"]["process_h5_path"],
            sub_region=(self.images_group[0], self.images_group[1], None, None, None, None) # (n_angles, n_z, n_x)
        )
        if expected_radios_shape[0] != self.images_group[1] - self.images_group[0]:
            msg = "Expected to find data with n_angles>=%d in %s but found n_angles=%d" % (
                self.images_group[1] - self.images_group[0],
                self.processing_options["read_chunk"]["process_file"],
                expected_radios_shape[0]
            )
            self.logger.fatal(msg)
            raise ValueError(msg)
        if expected_radios_shape[1] != self.dataset_info.radio_dims[1]:
            msg = "Expected to find data with n_y=%d in %s but found n_y=%d" % (
                self.dataset_info.radio_dims[1],
                self.processing_options["read_chunk"]["process_file"],
                expected_radios_shape[1]
            )
            self.logger.fatal(msg)
            raise ValueError(msg)


    def _init_reader_finalize(self):
        """
        Method called after _init_reader
        """
        self._check_resume_from_step()
        self.radios = self.chunk_reader.data
        self._radios_cropped = self.radios # compat. with ChunkedPipeline

    def _error_na(self):
        raise ValueError("This function is not applicable with the current class")


    def _do_nothing(self):
        pass

    _reset_sub_region = _error_na
    process_chunk = _error_na
    _process_chunk = _error_na
    _write_histogram = _do_nothing


    #
    # Pipeline initialization
    #

    def _init_pipeline(self):
        self._init_reader()
        self._init_flatfield()
        self._init_double_flatfield()
        self._init_ccd_corrections()
        self._init_radios_rotation()
        self._init_phase()
        self._init_unsharp()
        self._init_mlog()
        self._init_radios_movements()
        self._init_writer()
        self._configure_data_dumps()


    def _update_reader_configuration(self):
        """
        Modify self.processing_options["read_chunk"] to select a subset of the files
        """
        options = self.processing_options["read_chunk"]
        self._resume_from_step = options.get("step_name", None)
        if self._resume_from_step is not None:
            if self._resume_from_step == "sinogram":
                msg = "It makes no sense to use this class when resuming process from sinogram"
                self.logger.fatal(msg)
                raise ValueError(msg)
            return
        orig_options = self._orig_processing_options["read_chunk"]
        input_data_files = {}
        files_indices = sorted(orig_options["files"].keys())
        for i in range(self.images_group[0], self.images_group[1]):
            idx = files_indices[i]
            input_data_files[idx] = orig_options["files"][idx]
        options["files"] = input_data_files

    def _get_image_start_index(self):
        return self.images_group[0]

    @use_options("read_chunk", "chunk_reader")
    def _init_reader(self):
        if "read_chunk" not in self.processing_steps:
            raise ValueError("Cannot proceed without reading data")
        options = self.processing_options["read_chunk"]

        self._update_reader_configuration()

        if self._resume_from_step is None:
            # Standard case - start pipeline from raw data
            self.chunk_reader = ChunkReader(
                options["files"],
                sub_region=self.sub_region,
                convert_float=True,
                binning=options["binning"],
                dataset_subsampling=options["dataset_subsampling"]
            )
        else:
            # Resume pipeline from dumped intermediate step
            self.chunk_reader = HDF5Loader(
                options["process_file"],
                options["process_h5_path"],
                sub_region=(self.images_group[0], self.images_group[1], None, None, None, None),
                pre_allocate=True,
            )
        self._init_reader_finalize()


    def _update_flatfield_configuration(self):
        """
        Modify self.processing_options["flatfield"] to select a subset of the files
        """
        options = self.processing_options.get("flatfield", None)
        if options is None:
            return
        # ChunkReader.files accounts for subsampling
        options["projs_indices"] = sorted(self.chunk_reader.files_subsampled.keys())


    def _init_flatfield(self, shape=None):
        self._update_flatfield_configuration()
        ChunkedPipeline._init_flatfield(self, shape=shape)


    def _init_ctf_phase(self):
        options = self.processing_options["phase"]

        geo_pars_params = options["ctf_geo_pars"].copy()
        geo_pars_params["logger"] = self.logger
        geo_pars = GeoPars(**geo_pars_params)

        self.phase_retrieval = CTFPhaseRetrieval(
            geo_pars, options["delta_beta"],
            lim1=options["ctf_lim1"], lim2=options["ctf_lim2"],
            logger=self.logger
        )


    @use_options("phase", "phase_retrieval")
    def _init_phase(self):
        options = self.processing_options["phase"]
        if options["method"] == "CTF":
            self._init_ctf_phase()
        else:
            super()._init_phase()


    @use_options("save", "writer")
    def _init_writer(self):
        options = self.processing_options["save"]
        file_prefix = options["file_prefix"]
        output_dir = path.join(
            options["location"],
            file_prefix
        )
        file_prefix = "sinogram_" + file_prefix + str(
            "_%04d" % self.images_group[0]
        )
        entry = getattr(self.dataset_info.dataset_scanner, "entry", None)
        nx_info = {
            "process_name": "sinogram",
            "processing_index": 0,
            "config": {
                # "processing_options": self.processing_options, # Takes too much time to write, not useful for partial files
                "nabu_config": self.process_config.nabu_config
            },
            "entry": entry,
        }
        self._writer_configurator = WriterConfigurator(
            output_dir, file_prefix,
            file_format="hdf5",
            overwrite=options["overwrite"],
            start_index=self.images_group[0],
            logger=self.logger,
            nx_info=nx_info,
            write_histogram=False
        )
        self.writer = self._writer_configurator.writer
        self.histogram_writer = None
        self._writer_exec_args = self._writer_configurator._writer_exec_args
        self._writer_exec_kwargs = self._writer_configurator._writer_exec_kwargs


    def _get_cropped_radios(self):
        return self.radios


    def _configure_data_dumps(self):
        super()._configure_data_dumps()
        # dumping sinogram is natively done by this class.
        self._data_dump.pop("sinogram", None)


    #
    # Pipeline re-initialization
    #

    def _reset_images_group(self, images_group):
        if images_group[1] - images_group[0] != self.group_size:
            raise ValueError(
                "This class was instantiated with group_size=%d, but images_group=%s was passed"
                % (self.group_size, str(images_group))
            )
        self._set_group(images_group)
        # Reader
        self._update_reader_configuration()
        if self._resume_from_step is None:
            self.chunk_reader._set_files(self.processing_options["read_chunk"]["files"])
            self.chunk_reader._loaded = False
        else:
            self.chunk_reader._set_subregion((self.images_group[0], self.images_group[1], None, None, None, None))
            self.chunk_reader._loaded = False
        # Flat-field
        self._reset_flatfield()

    def _reset_flatfield(self):
        if self.flatfield is None:
            return
        new_radios_indices = self.processing_options["flatfield"]["projs_indices"]
        self.flatfield.radios_indices = np.array(new_radios_indices, dtype=np.int32)


    #
    # Pipeline execution
    #

    @pipeline_step("phase_retrieval", "Performing phase retrieval")
    def _retrieve_phase_ctf(self):
        options = self.processing_options["phase"]
        padding_mode = options["padding_type"]
        for i in range(self.radios.shape[0]):
            self.radios[i] = self.phase_retrieval.retrieve_phase(
                self.radios[i],
                padding_mode=padding_mode,
                # translation_vh=self.dataset_info.ctf_translations[i], # TODO unsupported
                translation_vh=None,
                normalize_by_mean=options["ctf_normalize_by_mean"],
            )


    def _retrieve_phase(self):
        if "phase" not in self.processing_steps:
            return
        options = self.processing_options["phase"]
        if options["method"] == "CTF":
            self._retrieve_phase_ctf()
        else:
            super()._retrieve_phase()


    def _process_group(self):
        self._flatfield()
        self._double_flatfield()
        self._ccd_corrections()
        self._retrieve_phase()
        self._apply_unsharp()
        self._take_log()
        self._rotate_projections()
        self._radios_movements()
        self._write_data(data=self.radios)


    def process_group(self, images_group=None):
        if images_group is not None:
            self._reset_images_group(images_group)
            self._reset_memory()
            self._init_writer()
            self._init_double_flatfield()
            self._configure_data_dumps()
        self._read_data()
        self._process_group()


class SinoStackPipeline:
    backend = "numpy"
    SinoNormalizationClass = ChunkedPipeline.SinoNormalizationClass
    SinoBuilderClass = ChunkedPipeline.SinoBuilderClass
    SinoDeringerClass = ChunkedPipeline.SinoDeringerClass
    FBPClass = ChunkedPipeline.FBPClass
    HistogramClass = ChunkedPipeline.HistogramClass


    """
    A pipeline for processing tomography data from sinograms.
    This class is to be used to complement GroupedPipeline.
    """
    def __init__(self, process_config, stack, projections, logger=None, extra_options=None):
        """
        Initialize a SinoStackPipeline.

        Parameters
        ----------
        process_config: ProcessConfig object
            Data structure with the processing configuration
        stack: tuple
            Tuple in the form (start_z, end_z)
        projections: str or array
           Pre-processed projections. Sinogram will be directly built from these projections.
           This parameter can of be the following type:

              - str: path to the HDF5 (master) file containing the projections (radios).
              - array: an array already containing the projections.


        Other parameters
        ----------------
        logger: Logger, optional
            Logging object
        extra_options: dict, optional
            Dictionary with extra options.
        """
        self.logger = LoggerOrPrint(logger)
        self._set_params(process_config, projections, stack, extra_options)
        self._init_pipeline()


    def _set_params(self, process_config, projections, stack, extra_options):
        self._set_extra_options(extra_options)
        self.process_config = process_config
        self.processing_steps = process_config.processing_steps
        self.processing_options = process_config.processing_options
        self.dataset_info = process_config.dataset_info
        self.projections = projections
        self._set_stack(stack)
        self._steps_name2component = {}
        self._steps_component2name = {}
        self._callbacks = {}
        self.use_radio_processing_margin = False
        self.chunk_reader = "get_projections"
        self._data_dump = {}


    def _set_stack(self, stack):
        if len(stack) != 2:
            raise ValueError("Expected 2-tuple (z_min, z_max)")
        self.z_min, self.z_max = stack
        self.images_stack = stack
        self.stack = stack


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "projections_processing_name": "sinogram",
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options


    def _get_shape(self, step_name):
        radios_3D_shape = self.radios.shape
        radio_2D_shape = radios_3D_shape[1:]
        if step_name == "sino_normalization":
            shape = radios_3D_shape
        elif step_name == "build_sino":
            shape = radios_3D_shape
        elif step_name == "sino_rings_correction":
            shape = self.sino_builder.output_shape
        elif step_name == "reconstruction":
            shape = self.sino_builder.output_shape[1:]
        else:
            raise ValueError("Cannot figure out shape for step '%s'" % step_name)
        self.logger.debug("Data shape for %s is %s" % (step_name, str(shape)))
        return shape

    #
    # Pipeline initialization
    #
    @use_options("read_chunk", "chunk_reader")
    def _init_reader(self):
        if isinstance(self.projections, str):
            fname = self.projections
            h5_entry = self.process_config.nabu_config["dataset"]["hdf5_entry"] or get_first_hdf5_entry(fname)
            self._data_h5_path = posixpath.join(
                h5_entry,
                self.extra_options["projections_processing_name"], "results", "data"
            )
            self.data_reader = HDF5Loader(
                fname,
                self._data_h5_path,
                sub_region=(None, None, self.stack[0], self.stack[1], None, None)
            )
            self.radios = self.data_reader.data
        elif isinstance(self.projections, np.ndarray):
            self.radios = self.projections
        else:
            raise ValueError("Expected str, list or array for 'projections'")


    def _get_slice_start_index(self):
        return self.z_min


    _allocate_array = ChunkedPipeline._allocate_array
    _allocate_sinobuilder_output = ChunkedPipeline._allocate_sinobuilder_output
    _prepare_reconstruction = ChunkedPipeline._prepare_reconstruction
    _allocate_recs = ChunkedPipeline._allocate_recs
    _init_sino_normalization = ChunkedPipeline._init_sino_normalization
    _init_sino_builder = ChunkedPipeline._init_sino_builder
    _init_sino_rings_correction = ChunkedPipeline._init_sino_rings_correction
    _init_histogram = ChunkedPipeline._init_histogram
    _init_writer = ChunkedPipeline._init_writer


    def _init_pipeline(self):
        self._init_reader()
        self._init_sino_normalization()
        self._init_sino_builder()
        self._init_sino_rings_correction()
        self._prepare_reconstruction()
        self._init_reconstruction()
        self._init_histogram()
        self._init_writer()


    #
    # Pipeline execution
    #

    @pipeline_step("chunk_reader", "Reading data")
    def _read_data(self):
        if isinstance(self.projections, str):
            self.data_reader.load_data()


    _normalize_sinos = ChunkedPipeline._normalize_sinos
    _build_sino = ChunkedPipeline._build_sino
    _destripe_sinos = ChunkedPipeline._destripe_sinos
    _init_reconstruction = ChunkedPipeline._init_reconstruction
    _reconstruct = ChunkedPipeline._reconstruct
    _compute_histogram = ChunkedPipeline._compute_histogram
    _write_data = ChunkedPipeline._write_data
    _write_histogram = ChunkedPipeline._write_histogram
    _get_process_name = ChunkedPipeline._get_process_name


    def _reset_stack(self, stack):
        self._set_stack(stack)
        self._init_reader()
        self._init_writer()


    def _process_stack(self):
        self._normalize_sinos()
        self._build_sino()
        self._destripe_sinos()
        self._reconstruct()
        self._compute_histogram()
        self._write_data()


    def process_stack(self, stack=None):
        if stack is not None:
            self._reset_stack(stack)
        self._read_data()
        self._process_stack()
